"""
=============
Heamap Slicer
=============

"""
import matplotlib.pyplot as plt
import numpy as np

from mpl_interactions import heatmap_slicer

x = np.linspace(0, np.pi, 100)
y = np.linspace(0, 10, 200)
X, Y = np.meshgrid(x, y)
data1 = np.sin(X) + np.exp(np.cos(Y))
data2 = np.cos(X) + np.exp(np.sin(Y))
fig, axes = heatmap_slicer(
    x,
    y,
    (data1, data2),
    slices="both",
    heatmap_names=("dataset 1", "dataset 2"),
    labels=("Some wild X variable", "Y axis"),
    interaction_type="move",
)
plt.tight_layout()

plt.show()



from pathlib import Path

import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Slider

import mpl_interactions.ipyplot as iplt
from mpl_interactions.controller import Controls


def test_eager_register():
    zoop = []

    def cb(**kwargs):
        zoop.append(1)

    ctrls = Controls(beep=(0, 1), boop=(0, 1))
    ctrls.register_callback(cb, None, eager=True)
    assert len(zoop) == 1

    ctrls = Controls(beep=(0, 1), boop=(0, 1))
    ctrls.register_callback(cb, "beep", eager=False)
    assert len(zoop) == 1

    ctrls.controls["beep"].set_val(0.5)
    assert len(zoop) == 2

    ctrls.controls["boop"].set_val(0.5)
    assert len(zoop) == 2


def test_save_animation(tmp_path: Path):
    # just test that no errors are thrown, don't actually check
    # that the saved animation makes any sense. Doing that seems hard :(
    x = np.linspace(0, np.pi, 100)
    tau = widgets.FloatSlider(min=1, max=10, step=1.5)

    def f1(x, amp, tau, beta):
        return amp * np.sin(x * tau) * x * beta

    def f2(x, amp, tau, beta):
        return amp * np.sin(x * beta) * x * tau

    x = np.linspace(0, 2 * np.pi, 200)

    fig, ax = plt.subplots()
    plt.subplots_adjust(bottom=0.25)
    axfreq = plt.axes([0.25, 0.1, 0.65, 0.03])
    amp_slider = Slider(axfreq, label="amp", valmin=0.05, valmax=10)
    ctrls = Controls(use_ipywidgets=True, tau=tau, beta=(1, 10, 15), amp=amp_slider)
    iplt.plot(x, f1, ax=ax, controls=ctrls)
    iplt.plot(x, f2, ax=ax, controls=ctrls)

    ctrls.save_animation(str(tmp_path / "animation-amp.gif"), fig, "amp", N_frames=10)
    ctrls.save_animation(str(tmp_path / "animation-beta.gif"), fig, "beta")
    ctrls.save_animation(str(tmp_path / "animation-tau.gif"), fig, "tau")


def test_scalar_inside_context():
    x = np.random.randn(20)
    y = 2 * x + np.random.randn(20) / 2

    x2 = np.random.randn(20)
    y2 = -2 * x2 - np.random.randn(20) / 2

    fig, ax = plt.subplots()
    ctrls = iplt.scatter(x, y, s=(10, 1000))
    # this shouldn't fail
    with ctrls:
        _ = iplt.scatter(x2, y2, s=ctrls["s"], ax=ax)



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure

from mpl_interactions.helpers import sca, update_datalim_from_bbox, update_datalim_from_xy


def test_bbox_update():
    fig, ax = plt.subplots()
    x = np.arange(10) / 10
    ax.plot(x)
    scat = ax.scatter([0], [0])
    x_scat = np.arange(3)
    y_scat_big = x_scat * 5
    y_scat_small = x_scat * 2
    scat.set_offsets(np.c_[x_scat, y_scat_big])
    update_datalim_from_bbox(ax, scat.get_datalim(ax.transData))
    assert ax.dataLim.bounds == (0, 0, 9, 10)
    scat.set_offsets(np.c_[x_scat, y_scat_small])
    update_datalim_from_bbox(ax, scat.get_datalim(ax.transData))
    assert ax.dataLim.bounds == (0, 0, 9, 10)
    scat.set_offsets(np.c_[x_scat, y_scat_big])
    update_datalim_from_bbox(ax, scat.get_datalim(ax.transData))
    assert ax.dataLim.bounds == (0, 0, 9, 10)
    scat.set_offsets(np.c_[x_scat, y_scat_small])
    update_datalim_from_bbox(ax, scat.get_datalim(ax.transData), stretch_y=False)
    assert ax.dataLim.bounds == (0.0, 0.0, 9.0, 4.0)


def test_xy_update():
    fig, ax = plt.subplots()
    x = np.arange(10) / 10
    ax.plot(x)
    scat = ax.scatter([0], [0])
    x_scat = np.arange(3)
    y_scat_big = x_scat * 5
    y_scat_small = x_scat * 2
    update_datalim_from_xy(ax, x_scat, y_scat_big)
    assert ax.dataLim.bounds == (0, 0, 9, 10)
    scat.set_offsets(np.c_[x_scat, y_scat_small])
    update_datalim_from_xy(ax, x_scat, y_scat_small)
    assert ax.dataLim.bounds == (0, 0, 9, 10)
    update_datalim_from_xy(ax, x_scat, y_scat_big)
    assert ax.dataLim.bounds == (0, 0, 9, 10)
    update_datalim_from_xy(ax, x_scat, y_scat_small, stretch_y=False)
    assert ax.dataLim.bounds == (0.0, 0.0, 9.0, 4.0)


def test_sca():
    fig = Figure()
    ax = fig.add_subplot(111)

    # this shouldn't fail
    sca(ax)



__all__ = [
    "set_param_values",
]


def set_param_values(controls, params):
    """
    A bit sketch because it set the index for ipywidgets
    and direct value for mpl sliders. big TODO there.
    """
    for p, v in params.items():
        slider = controls.controls[p]
        if "Box" in str(slider.__class__):
            for obj in slider.children:
                if "Slider" in str(obj.__class__):
                    obj.value = v
        else:
            slider.set_val(v)



from .pyplot import interactive_axhline as axhline
from .pyplot import interactive_axvline as axvline
from .pyplot import interactive_hist as hist
from .pyplot import interactive_imshow as imshow
from .pyplot import interactive_plot as plot
from .pyplot import interactive_scatter as scatter
from .pyplot import interactive_text as text
from .pyplot import interactive_title as title
from .pyplot import interactive_xlabel as xlabel
from .pyplot import interactive_ylabel as ylabel



"""Functions that will be useful irrespective of backend."""

from collections.abc import Callable

import numpy as np
from matplotlib import __version__ as mpl_version
from matplotlib import get_backend
from matplotlib.colors import TABLEAU_COLORS, XKCD_COLORS, to_rgba_array
from matplotlib.path import Path
from matplotlib.pyplot import close, ioff, subplots
from matplotlib.widgets import LassoSelector
from numpy import asanyarray, asarray, max, min

from .controller import gogogo_controls, prep_scalars
from .helpers import (
    callable_else_value_no_cast,
    create_slider_format_dict,
    gogogo_figure,
    notebook_backend,
)
from .utils import figure, nearest_idx
from .xarray_helpers import get_hs_axes, get_hs_extent, get_hs_fmts

# functions that are methods
__all__ = [
    "heatmap_slicer",
    "zoom_factory",
    "panhandler",
    "image_segmenter",
    "hyperslicer",
]


def heatmap_slicer(
    X,
    Y,
    heatmaps,
    slices="horizontal",
    heatmap_names=None,
    linecolor="k",
    labels=("X", "Y"),
    interaction_type="move",
    fig=None,
    figsize=(18, 9),
    **pcolormesh_kwargs,
):

    """
    Compare horizontal and/or vertical slices across multiple arrays.

    Parameters
    ----------
    X,Y : 1D array
    heatmaps : array_like
        must be 2-D or 3-D. If 3-D the last two axes should be (X,Y)
    slices : {'horizontal', 'vertical', 'both'}
        Direction to draw slice on heatmap. both will draw horizontal and vertical traces on
        the same plot, while both_separate will make a line plot for each.
    heatmap_names : (String, String, ...)
        An iterable with the names of the heatmaps. If provided it must have as many names
        as there are heatmaps
    figsize : tuple of number, default: (18, 9)
        The size of the created figure. Ignored if *fig* is not None.
    linecolor : colorlike, default: 'k'
        The color of the cursor showing the slices. Must be a valid Matplotlib linecolor.
    labels : (string, string), default: ("X", "Y")
        The labels for the x and y axes.
    interaction_type : str
        Update on mouse movement or mouse click. Options are {'move','click'}
    fig : matplotlib figure, optional
        The figure to use for the heatmap_slicer. Useful when embedding into a gui.
        If you are embedding into a gui make sure you set up the gui canvas first
        and then pass the figure to this function
    **pcolormesh_kwargs
        kwargs passed to ``ax.pcolormesh``.

    Returns
    -------
    fig : matplotlib.Figure.figure
    ax : tuple of matplotlib.axes.Axes
    """
    horiz = vert = False
    if slices == "both":
        num_line_axes = 2
        horiz_axis = -2
        vert_axis = -1
        horiz = vert = True
    else:
        horiz_axis = -1
        vert_axis = -1
        num_line_axes = 1
        if slices == "horizontal":
            horiz = True
        elif slices == "vertical":
            vert = True
        else:
            raise ValueError("Valid options for slices are {horizontal, vertical, both}")

    heatmaps = asarray(heatmaps)
    if heatmap_names is None:
        heatmap_names = [f"heatmap_{i}" for i in range(heatmaps.shape[0])]

    if heatmaps.ndim == 3:
        num_axes = num_line_axes + heatmaps.shape[0]
        if type(heatmap_names) is str or (len(heatmap_names) != heatmaps.shape[0]):
            raise ValueError("need to provide at least as many heatmap_names as heatmaps")
    elif heatmaps.ndim == 2:
        heatmaps = heatmaps.reshape(1, *heatmaps.shape)
        if type(heatmap_names) is str:
            heatmap_names = [heatmap_names]
        num_axes = num_line_axes + 1
    else:
        raise ValueError(f"heatmaps must be 2D or 3D but is {heatmaps.ndim}D")

    if fig is None:
        fig, axes = subplots(1, num_axes, figsize=figsize)
    else:
        axes = fig.subplots(1, num_axes)

    hlines = []
    vlines = []
    init_idx = 0
    axes[0].set_ylabel(labels[1])
    X = asarray(X)
    Y = asarray(Y)
    # mpl pcolormesh from version 3.3+ handles len(X), len(Y) equal to Z shape
    # differently than <2. (Unquestionably better, but different enough to justify a shim)
    # https://github.com/matplotlib/matplotlib/pull/16258
    shading = pcolormesh_kwargs.pop("shading", "auto")

    x_centered = X[:-1] + (X[1:] - X[:-1]) / 2
    y_centered = Y[:-1] + (Y[1:] - Y[:-1]) / 2
    for i, ax in enumerate(axes[:-num_line_axes]):
        ax.pcolormesh(X, Y, heatmaps[i], shading=shading, **pcolormesh_kwargs)
        ax.set_xlabel(labels[0])
        ax.set_title(heatmap_names[i])
        hmap_shape = asanyarray(heatmaps[i]).shape

        if i > 0:
            ax.set_yticklabels([])
        if horiz:
            same_shape = X.shape[0] == hmap_shape[1]
            if same_shape:
                x = X
            else:
                x = x_centered
            data_line = axes[horiz_axis].plot(
                x, heatmaps[i, init_idx, :], label=f"{heatmap_names[i]}"
            )[0]
            hlines.append((same_shape, ax.axhline(Y[init_idx], color=linecolor), data_line))

        if vert:
            same_shape = Y.shape[0] == hmap_shape[0]
            if same_shape:
                y = Y
            else:
                y = y_centered
            data_line = axes[vert_axis].plot(
                y, heatmaps[i, :, init_idx], label=f"{heatmap_names[i]}"
            )[0]
            vlines.append((same_shape, ax.axvline(X[init_idx], color=linecolor), data_line))

    minimum = min(heatmaps)
    maximum = max(heatmaps)
    if vert:
        axes[vert_axis].set_title("Vertical")
        axes[vert_axis].set_ylim([minimum, maximum])
        axes[vert_axis].legend()
    if horiz:
        axes[horiz_axis].set_title("Horizontal")
        axes[horiz_axis].set_ylim([minimum, maximum])
        axes[horiz_axis].legend()

    def _gen_idxs(orig, centered, same_shape, event_data):
        """
        is there a better way? probably, but this gets the job done
        so here we are...
        """
        if same_shape:
            data_idx = nearest_idx(orig, event_data)
            disp_idx = nearest_idx(orig, event_data)
            arr = orig
        else:
            disp_idx = nearest_idx(centered, event_data)
            data_idx = nearest_idx(centered, event_data)
            arr = centered
        return arr, data_idx, disp_idx

    def update_lines(event):
        if event.inaxes in axes[:-num_line_axes]:
            y = None
            for i, (same_shape, display_line, data_line) in enumerate(hlines):
                if y is None:
                    y, data_idx, disp_idx = _gen_idxs(Y, y_centered, same_shape, event.ydata)
                display_line.set_ydata(y[disp_idx])
                data_line.set_ydata(heatmaps[i, data_idx])
            x = None
            for i, (same_shape, display_line, data_line) in enumerate(vlines):
                if x is None:
                    x, data_idx, disp_idx = _gen_idxs(X, x_centered, same_shape, event.xdata)
                display_line.set_xdata(x[disp_idx])
                data_line.set_ydata(heatmaps[i, :, data_idx])
        fig.canvas.draw_idle()

    if interaction_type == "move":
        fig.canvas.mpl_connect("motion_notify_event", update_lines)
    elif interaction_type == "click":
        fig.canvas.mpl_connect("button_press_event", update_lines)
    else:
        close(fig)
        raise ValueError(
            f"{interaction_type} is not a valid option for interaction_type, valid options are 'click' or 'move'"
        )
    return fig, axes


# based on https://gist.github.com/tacaswell/3144287
def zoom_factory(ax, base_scale=1.1):
    """
    Add ability to zoom with the scroll wheel.

    Parameters
    ----------
    ax : matplotlib axes object
        axis on which to implement scroll to zoom
    base_scale : float
        how much zoom on each tick of scroll wheel

    Returns
    -------
    disconnect_zoom : function
        call this to disconnect the scroll listener
    """

    def limits_to_range(lim):
        return lim[1] - lim[0]

    fig = ax.get_figure()  # get the figure of interest
    fig.canvas.capture_scroll = True
    has_toolbar = hasattr(fig.canvas, "toolbar") and fig.canvas.toolbar is not None
    if has_toolbar:
        # it might be possible to have an interactive backend without
        # a toolbar. I'm not sure so being safe here
        toolbar = fig.canvas.toolbar
        toolbar.push_current()
    orig_xlim = ax.get_xlim()
    orig_ylim = ax.get_ylim()
    orig_yrange = limits_to_range(orig_ylim)
    orig_xrange = limits_to_range(orig_xlim)
    orig_center = ((orig_xlim[0] + orig_xlim[1]) / 2, (orig_ylim[0] + orig_ylim[1]) / 2)

    def zoom_fun(event):
        if event.inaxes is not ax:
            return
        # get the current x and y limits
        cur_xlim = ax.get_xlim()
        cur_ylim = ax.get_ylim()
        # set the range
        (cur_xlim[1] - cur_xlim[0]) * 0.5
        (cur_ylim[1] - cur_ylim[0]) * 0.5
        xdata = event.xdata  # get event x location
        ydata = event.ydata  # get event y location
        if event.button == "up":
            # deal with zoom in
            scale_factor = base_scale
        elif event.button == "down":
            # deal with zoom out
            scale_factor = 1 / base_scale
        else:
            # deal with something that should never happen
            scale_factor = 1
        # set new limits
        new_xlim = [
            xdata - (xdata - cur_xlim[0]) / scale_factor,
            xdata + (cur_xlim[1] - xdata) / scale_factor,
        ]
        new_ylim = [
            ydata - (ydata - cur_ylim[0]) / scale_factor,
            ydata + (cur_ylim[1] - ydata) / scale_factor,
        ]
        new_yrange = limits_to_range(new_ylim)
        new_xrange = limits_to_range(new_xlim)

        if abs(new_yrange) > abs(orig_yrange):
            new_ylim = orig_center[1] - new_yrange / 2, orig_center[1] + new_yrange / 2
        if abs(new_xrange) > abs(orig_xrange):
            new_xlim = orig_center[0] - new_xrange / 2, orig_center[0] + new_xrange / 2
        ax.set_xlim(new_xlim)
        ax.set_ylim(new_ylim)

        if has_toolbar:
            toolbar.push_current()
        ax.figure.canvas.draw_idle()  # force re-draw

    # attach the call back
    cid = fig.canvas.mpl_connect("scroll_event", zoom_fun)

    def disconnect_zoom():
        fig.canvas.mpl_disconnect(cid)

    # return the disconnect function
    return disconnect_zoom


class panhandler:
    """
    Enable panning a plot with any mouse button.

    .. code-block:: python

       handler = panhandler(my_figure)

       # Let it be disabled and garbage collected
       handler = None

    Parameters
    ----------
    button : int
        Determines which button will be used (default right click).
        Left: 1
        Middle: 2
        Right: 3
    """

    def __init__(self, fig, button=3):
        self.fig = fig
        self._id_drag = None
        self.button = button
        self._id_press = None
        self._id_release = None

        self.enable()

    @property
    def enabled(self) -> bool:
        """
        Status of the panhandler, whether it's enabled or disabled.
        """
        return self._id_press is not None and self._id_release is not None

    def enable(self):
        """
        Enable the panhandler. It should not be necessary to call this function
        unless it's used after a call to :meth:`panhandler.disable`.

        Raises
        ------
        RuntimeError
            If the panhandler is already enabled.
        """
        if self.enabled:
            raise RuntimeError("The panhandler is already enabled")

        self._id_press = self.fig.canvas.mpl_connect("button_press_event", self.press)
        self._id_release = self.fig.canvas.mpl_connect("button_release_event", self.release)

    def disable(self):
        """
        Disable the panhandler.

        Raises
        ------
        RuntimeError
            If the panhandler is already disabled.
        """
        if not self.enabled:
            raise RuntimeError("The panhandler is already disabled")

        self.fig.canvas.mpl_disconnect(self._id_press)
        self.fig.canvas.mpl_disconnect(self._id_release)

        self._id_press = None
        self._id_release = None

    def _cancel_action(self):
        self._xypress = []
        if self._id_drag:
            self.fig.canvas.mpl_disconnect(self._id_drag)
            self._id_drag = None

    def press(self, event):
        if event.button != self.button:
            self._cancel_action()
            return

        x, y = event.x, event.y

        self._xypress = []
        for i, a in enumerate(self.fig.get_axes()):
            if (
                x is not None
                and y is not None
                and a.in_axes(event)
                and a.get_navigate()
                and a.can_pan()
            ):
                a.start_pan(x, y, event.button)
                self._xypress.append((a, i))
                self._id_drag = self.fig.canvas.mpl_connect("motion_notify_event", self._mouse_move)

    def release(self, event):
        self._cancel_action()
        self.fig.canvas.mpl_disconnect(self._id_drag)

        for a, _ind in self._xypress:
            a.end_pan()
        if not self._xypress:
            self._cancel_action()
            return
        self._cancel_action()

    def _mouse_move(self, event):
        for a, _ind in self._xypress:
            # safer to use the recorded button at the _press than current
            # button: # multiple button can get pressed during motion...
            a.drag_pan(1, event.key, event.x, event.y)
        self.fig.canvas.draw_idle()


class image_segmenter:
    """
    Manually segment an image with the lasso selector.
    """

    def __init__(
        self,
        img,
        nclasses=1,
        mask=None,
        mask_colors=None,
        mask_alpha=0.75,
        lineprops=None,
        props=None,
        lasso_mousebutton="left",
        pan_mousebutton="middle",
        ax=None,
        figsize=(10, 10),
        **kwargs,
    ):
        """
        Create an image segmenter. Any ``kwargs`` will be passed through to the ``imshow``
        call that displays *img*.

        Parameters
        ----------
        img : array_like
            A valid argument to imshow
        nclasses : int, default 1
        mask : arraylike, optional
            If you want to pre-seed the mask
        mask_colors : None, color, or array of colors, optional
            the colors to use for each class. Unselected regions will always be totally transparent
        mask_alpha : float, default .75
            The alpha values to use for selected regions. This will always override the alpha values
            in mask_colors if any were passed
        lineprops : dict, default: None
            DEPRECATED - use props instead.
            lineprops passed to LassoSelector. If None the default values are:
            {"color": "black", "linewidth": 1, "alpha": 0.8}
        props : dict, default: None
            props passed to LassoSelector. If None the default values are:
            {"color": "black", "linewidth": 1, "alpha": 0.8}
        lasso_mousebutton : str, or int, default: "left"
            The mouse button to use for drawing the selecting lasso.
        pan_mousebutton : str, or int, default: "middle"
            The button to use for `~mpl_interactions.generic.panhandler`. One of 'left', 'middle' or
            'right', or 1, 2, 3 respectively.
        ax : `matplotlib.axes.Axes`, optional
            The axis on which to plot. If *None* a new figure will be created.
        figsize : (float, float), optional
            passed to plt.figure. Ignored if *ax* is given.
        **kwargs
            All other kwargs will passed to the imshow command for the image
        """
        # ensure mask colors is iterable and the same length as the number of classes
        # choose colors from default color cycle?

        self.mask_alpha = mask_alpha

        if mask_colors is None:
            # this will break if there are more than 10 classes
            if nclasses <= 10:
                self.mask_colors = to_rgba_array(list(TABLEAU_COLORS)[:nclasses])
            else:
                # up to 949 classes. Hopefully that is always enough....
                self.mask_colors = to_rgba_array(list(XKCD_COLORS)[:nclasses])
        else:
            self.mask_colors = to_rgba_array(np.atleast_1d(mask_colors))
            # should probably check the shape here
        self.mask_colors[:, -1] = self.mask_alpha

        self._img = np.asarray(img)

        if mask is None:
            self.mask = np.zeros(self._img.shape[:2])
            """See :doc:`/examples/image-segmentation`."""
        else:
            self.mask = mask

        self._overlay = np.zeros((*self._img.shape[:2], 4))
        self.nclasses = nclasses
        for i in range(nclasses + 1):
            idx = self.mask == i
            if i == 0:
                self._overlay[idx] = [0, 0, 0, 0]
            else:
                self._overlay[idx] = self.mask_colors[i - 1]
        if ax is not None:
            self.ax = ax
            self.fig = self.ax.figure
        else:
            with ioff():
                self.fig = figure(figsize=figsize)
                self.ax = self.fig.gca()
        self.displayed = self.ax.imshow(self._img, **kwargs)
        self._mask = self.ax.imshow(self._overlay)

        default_props = {"color": "black", "linewidth": 1, "alpha": 0.8}
        if (props is None) and (lineprops is None):
            props = default_props
        elif (lineprops is not None) and (mpl_version >= "3.7"):
            print("*lineprops* is deprecated - please use props")
            props = {"color": "black", "linewidth": 1, "alpha": 0.8}

        useblit = False if "ipympl" in get_backend().lower() else True
        button_dict = {"left": 1, "middle": 2, "right": 3}
        if isinstance(pan_mousebutton, str):
            pan_mousebutton = button_dict[pan_mousebutton.lower()]
        if isinstance(lasso_mousebutton, str):
            lasso_mousebutton = button_dict[lasso_mousebutton.lower()]

        if mpl_version < "3.7":
            self.lasso = LassoSelector(
                self.ax, self._onselect, lineprops=props, useblit=useblit, button=lasso_mousebutton
            )
        else:
            self.lasso = LassoSelector(
                self.ax, self._onselect, props=props, useblit=useblit, button=lasso_mousebutton
            )
        self.lasso.set_visible(True)

        pix_x = np.arange(self._img.shape[0])
        pix_y = np.arange(self._img.shape[1])
        xv, yv = np.meshgrid(pix_y, pix_x)
        self.pix = np.vstack((xv.flatten(), yv.flatten())).T

        self.ph = panhandler(self.fig, button=pan_mousebutton)
        self.disconnect_zoom = zoom_factory(self.ax)
        self.current_class = 1
        self.erasing = False

    def _onselect(self, verts):
        self.verts = verts
        p = Path(verts)
        self.indices = p.contains_points(self.pix, radius=0).reshape(self.mask.shape)
        if self.erasing:
            self.mask[self.indices] = 0
            self._overlay[self.indices] = [0, 0, 0, 0]
        else:
            self.mask[self.indices] = self.current_class
            self._overlay[self.indices] = self.mask_colors[self.current_class - 1]

        self._mask.set_data(self._overlay)
        self.fig.canvas.draw_idle()

    def _ipython_display_(self):
        display(self.fig.canvas)  # noqa: F405, F821


def hyperslicer(
    arr,
    cmap=None,
    norm=None,
    aspect=None,
    interpolation=None,
    alpha=None,
    vmin=None,
    vmax=None,
    vmin_vmax=None,
    origin=None,
    extent=None,
    autoscale_cmap=True,
    filternorm=True,
    filterrad=4.0,
    resample=None,
    url=None,
    ax=None,
    slider_formats=None,
    title=None,
    force_ipywidgets=False,
    play_buttons=False,
    is_color_image=False,
    controls=None,
    display_controls=True,
    **kwargs,
):

    """
    View slices from a hyperstack of images selected by sliders. Also accepts Xarray.DataArrays
    in which case the axes names and coordinates will be inferred from the xarray dims and coords.

    Parameters
    ----------
    arr : arraylike or xarray
        Hyperstack of images. The last 2 or 3 dimensions will be treated as individiual images.
        If an xarray.DataArray then the dimensions will be automatically inferred.
    cmap : str or `~matplotlib.colors.Colormap`
        The Colormap instance or registered colormap name used to map
        scalar data to colors. This parameter is ignored for RGB(A) data.
        forwarded to matplotlib
    norm : `~matplotlib.colors.Normalize`, optional
        The `~matplotlib.colors.Normalize` instance used to scale scalar data to the [0, 1]
        range before mapping to colors using *cmap*. By default, a linear
        scaling mapping the lowest value to 0 and the highest to 1 is used.
        This parameter is ignored for RGB(A) data.
        forwarded to matplotlib
    autoscale_cmap : bool
        If True rescale the colormap for every function update. Will not update
        if vmin and vmax are provided or if the returned image is RGB(A) like.
        forwarded to matplotlib
    aspect : {'equal', 'auto'} or float
        forwarded to matplotlib
        interpolation : str
        forwarded to matplotlib
    ax : matplotlib axis, optional
        if None a new figure and axis will be created
    slider_formats : None, string, or dict
        If None a default value of decimal points will be used. Uses the new {} style formatting
    title : None or string
        If a string then you can have it update automatically using string formatting of the names
        of the parameters. i.e. to include the current value of tau: title='the value of tau is: {tau:.2f}'
    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    is_color_image : boolean
        If True, will treat the last 3 dimensions as comprising a color images and will only set up sliders for the first arr.ndim - 3 dimensions.
    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    display_controls : boolean
        Whether the controls should display on creation. Ignored if controls is specified.

    Returns
    -------
    controls
    """

    arr = np.squeeze(arr)

    arr_type = "numpy"
    if "xarray.core.dataarray.DataArray" in str(arr.__class__):
        arr_type = "xarray"
    elif "dask.array.core.Array" in str(arr.__class__):
        arr_type = "dask"

    if arr.ndim < 3 + is_color_image:
        raise ValueError(
            f"arr must be at least {3+is_color_image}D but it is {arr.ndim}D. mpl_interactions.imshow for 2D images."
        )

    if is_color_image:
        im_dims = 3
    else:
        im_dims = 2

    ipympl = notebook_backend()
    fig, ax = gogogo_figure(ipympl, ax)
    ipympl or force_ipywidgets
    slider_format_strings = create_slider_format_dict(slider_formats)

    name_to_dim = {}
    slices = [0 for i in range(arr.ndim - im_dims)]

    names = None
    axes = None
    if arr_type != "xarray":
        if "names" in kwargs:
            names = kwargs.pop("names")

        elif "axes" in kwargs:
            axes = kwargs.pop("axes")

    else:
        axes = get_hs_axes(arr, is_color_image=is_color_image)

    # Just pass in an array - no kwargs
    for i in range(arr.ndim - im_dims):

        start, stop = None, None
        name = f"axis{i}"
        if name in kwargs:
            if len(kwargs[name]) == 2:
                start, stop = kwargs.pop(name)
            else:
                kwargs.pop(name)

        if axes is not None and axes[i] is not None:
            # now we assume the axes[i] has one of the following forms
            # ('mu', (0,1))
            # ('mu', np.array)
            # ('mu', 0, 1)
            # (0, 1)
            # 'mu'
            # np.array or a list
            a = axes[i]
            if isinstance(a, str):
                # axes = ['mu', ]
                name = a
            elif isinstance(a, tuple):
                if len(a) == 3:
                    # axes = [('mu', 0, 1)]
                    name = a[0]
                    kwargs[name] = (*a[1:], arr.shape[i])
                elif len(a) == 2:
                    if isinstance(a[0], str):
                        # axes = [('mu', (0,1))]
                        # axes = [('mu', np.linspace())]
                        # axes = [('mu', {('type1', 'type2', 'type3')}]
                        name = a[0]
                        if isinstance(a[1], tuple) or (isinstance(a[1], list) and len(a[1]) == 2):
                            kwargs[name] = (*a[1], arr.shape[i])
                        elif isinstance(a[1], np.ndarray) or isinstance(a[1], list):
                            kwargs[name] = a[1]
                        elif isinstance(a[1], set):
                            kwargs[name] = a[1]
                    elif np.isscalar(a[0]) and np.isscalar(a[1]):
                        # axes = [(0,1)]
                        kwargs[name] = (a[0], a[1], arr.shape[i])
                        slider_format_strings[name] = "{:.0f}"
            elif isinstance(a, list) or isinstance(np.ndarray):
                # no name only values
                kwargs[name] = a
        elif names is not None and names[i] is not None:
            name = names[i]
        name_to_dim[name] = i

        if name not in kwargs:
            slider_format_strings[name] = "{:.0f}"
            kwargs[name] = np.arange(arr.shape[i])

    if arr_type == "xarray":
        slider_format_strings = get_hs_fmts(arr, is_color_image=is_color_image)
        if extent is None:
            extent = get_hs_extent(arr, is_color_image=is_color_image)
    else:
        if "extent" not in kwargs:
            extent = None

    extra_ctrls = []
    funcs, extra_ctrls, param_excluder = prep_scalars(kwargs, vmin=vmin, vmax=vmax, alpha=alpha)
    vmin = funcs["vmin"]
    vmax = funcs["vmax"]
    alpha = funcs["alpha"]
    if vmin_vmax is not None:
        if isinstance(vmin_vmax, tuple) and not isinstance(vmin_vmax[0], str):
            vmin_vmax = ("r", *vmin_vmax)
        kwargs["vmin_vmax"] = vmin_vmax

    controls, params = gogogo_controls(
        kwargs,
        controls,
        display_controls,
        slider_format_strings,
        play_buttons,
        extra_ctrls,
        allow_dupes=True,
    )
    if vmin_vmax is not None:
        params.pop("vmin_vmax")
        params["vmin"] = controls.params["vmin"]
        params["vmax"] = controls.params["vmax"]

        def vmin(**kwargs):
            return kwargs["vmin"]

        def vmax(**kwargs):
            return kwargs["vmax"]

    def update(params, indices, cache):
        if title is not None:
            ax.set_title(title.format(**params))

        for k, v in indices.items():
            try:
                slices[name_to_dim[k]] = v
            except KeyError:
                # this is necessary to allow things
                # like vmax = (240, 250)
                pass

        new_data = arr[tuple(slices)]
        im.set_data(new_data)

        if autoscale_cmap and (new_data.ndim != 3) and vmin is None and vmax is None:
            im.norm.autoscale(new_data)

        if isinstance(vmin, Callable):
            im.norm.vmin = vmin(**param_excluder(params, "vmin"))
        if isinstance(vmax, Callable):
            im.norm.vmax = vmax(**param_excluder(params, "vmax"))
        if isinstance(alpha, Callable):
            im.set_alpha(callable_else_value_no_cast(alpha, param_excluder(params, "alpha"), cache))

    controls._register_function(update, fig, params.keys())
    # make it once here so we can use the dims in update
    new_data = arr[tuple(0 for i in range(arr.ndim - im_dims))]
    im = ax.imshow(
        new_data,
        cmap=cmap,
        norm=norm,
        aspect=aspect,
        interpolation=interpolation,
        alpha=alpha,
        vmin=callable_else_value_no_cast(vmin, params),
        vmax=callable_else_value_no_cast(vmax, params),
        origin=origin,
        extent=extent,
        filternorm=filternorm,
        filterrad=filterrad,
        resample=resample,
        url=url,
    )
    # this is necessary to make calls to plt.colorbar behave as expected
    # i know it's bad news to use private methods :(
    # but idk how else to accomplish being a psuedo-pyplot
    ax._sci(im)
    if title is not None:
        ax.set_title(title.format(**params))

    return controls



from os.path import dirname, realpath

import matplotlib.pyplot as plt
import numpy as np

from mpl_interactions.generic import heatmap_slicer, image_segmenter


# just smoketests here. hadn't set image comparison styling properly
def test_heatmap_slicer():
    x = np.linspace(0, np.pi, 100)
    y = np.linspace(0, 10, 200)
    X, Y = np.meshgrid(x, y)
    data1 = np.sin(X) + np.exp(np.cos(Y))
    data2 = np.cos(X) + np.exp(np.sin(Y))
    fig, axes = heatmap_slicer(
        x,
        y,
        (data1, data2),
        slices="both",
        heatmap_names=("dataset 1", "dataset 2"),
        labels=("Some wild X variable", "Y axis"),
        interaction_type="move",
        cmap="plasma",
    )


def test_image_segmentation():
    image = plt.imread(
        "https://github.com/matplotlib/matplotlib/raw/v3.3.0/lib/matplotlib/mpl-data/sample_data/ada.png"
    )
    script_dir = realpath(dirname(__file__))
    mask = np.load(f"{script_dir}/../docs/examples/ada-mask.npy")
    preloaded = image_segmenter(image, nclasses=3, mask=mask)

    return preloaded.fig



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import __version__ as mpl_version
from matplotlib.testing.decorators import check_figures_equal
from packaging import version

import mpl_interactions.ipyplot as iplt
from mpl_interactions.pyplot import interactive_plot

from ._util import set_param_values

np.random.seed(1111111121)

mpl_gr_32 = version.parse(mpl_version) >= version.parse("3.3")


def f_hist(loc, scale):
    return np.random.randn(1000) * scale + loc


# # smoke test
# def test_hist_plot(fig_test, fig_ref):
#     fig, ax =
#     test_ax = fig_test.add_subplot()
#     _ = interactive_hist(f_hist, density=True, loc=(5.5, 100), scale=(10, 15), ax=test_ax)


def f1(x, tau, beta):
    return np.sin(x * tau) * x * beta


def f2(x, tau, beta):
    return np.sin(x * beta) * x * tau


def f1(x, tau, beta):
    return np.sin(x * tau) * x * beta


def f2(x, tau, beta):
    return np.sin(x * beta) * x * tau


x = np.linspace(0, np.pi, 100)
tau = (5, 10, 100)
beta = (1, 2)


def test_multiple_functions():
    fig, ax = plt.subplots()
    controls = interactive_plot(x, f1, tau=tau, beta=beta, label="f1")
    interactive_plot(x, f2, controls=controls, label="f2")
    _ = plt.legend()
    return fig


@check_figures_equal(extensions=["png"])
def test_plot(fig_test, fig_ref):
    test_ax = fig_test.add_subplot()

    # TODO: fix the horrible ylim scaling
    # get it outside of the plot command
    ylims = (-10, 10)
    controls = interactive_plot(x, f1, beta=beta, tau=tau, label="f1", ax=test_ax, ylim=ylims)
    interactive_plot(
        x, f2, controls=controls, label="custom label!", linestyle="--", ax=test_ax, ylim=ylims
    )
    set_param_values(controls, {"beta": 5, "tau": 4})
    test_ax.legend()

    ref_ax = fig_ref.add_subplot()
    ref_ax.plot(x, f1(x, **controls.params), label="f1")
    ref_ax.plot(x, f2(x, **controls.params), linestyle="--", label="custom label!")
    ref_ax.legend()
    ref_ax.set_ylim(ylims)
    for fig in controls.control_figures:
        plt.close(fig)


@check_figures_equal(extensions=["png"])
def test_imshow_scalars(fig_test, fig_ref):
    # and by proxy all the scalar handling

    mask_arr = np.random.randn(10, 10) * 10

    def mask(min_distance):
        new_arr = np.copy(mask_arr)
        new_arr[mask_arr < min_distance] = 0
        return new_arr

    test_ax = fig_test.add_subplot()
    ctrls = iplt.imshow(mask, min_distance=(1, 10), alpha=(0, 1), ax=test_ax)
    set_param_values(ctrls, {"min_distance": 5.5, "alpha": 0.75})
    ref_ax = fig_ref.add_subplot()
    print(ctrls.params)
    ref_ax.imshow(mask(ctrls.params["min_distance"]), alpha=ctrls.params["alpha"])
    for fig in ctrls.control_figures:
        plt.close(fig)


def test_title():
    fig, ax = plt.subplots()
    ctrls = iplt.plot(1, 1, E=3e7)
    expected = "E=3.00e+07"
    iplt.title("E={E:.2e}", controls=ctrls)
    assert not isinstance(ctrls.params["E"], np.ndarray)
    assert ax.get_title() == expected
    plt.close()
    fig, ax = plt.subplots()
    ctrls = iplt.plot(1, 1, E=np.array([3e7, 3e9]))
    iplt.title("E={E:.2e}", controls=ctrls)
    assert ax.get_title() == expected
    plt.close()



"""Control the output of standard plotting functions such as :func:`~matplotlib.pyplot.plot` and
:func:`~matplotlib.pyplot.hist` using sliders and other widgets. When using the ``ipympl`` backend
these functions will leverage ipywidgets for the controls, otherwise they will use the built-in
Matplotlib widgets."""


from collections.abc import Callable
from numbers import Number

import matplotlib.markers as mmarkers
import numpy as np
from matplotlib.collections import PatchCollection
from matplotlib.colors import to_rgba_array
from matplotlib.patches import Rectangle

from .controller import gogogo_controls, prep_scalars
from .helpers import (
    callable_else_value,
    callable_else_value_no_cast,
    create_slider_format_dict,
    eval_xy,
    gogogo_figure,
    notebook_backend,
    sca,
    update_datalim_from_bbox,
)
from .mpl_kwargs import (
    Line2D_kwargs_list,
    Text_kwargs_list,
    collection_kwargs_list,
    imshow_kwargs_list,
    kwarg_popper,
)

__all__ = [
    "interactive_plot",
    "interactive_hist",
    "interactive_scatter",
    "interactive_imshow",
    "interactive_axhline",
    "interactive_axvline",
    "interactive_title",
    "interactive_xlabel",
    "interactive_ylabel",
    "interactive_text",
]


def interactive_plot(
    *args,
    parametric=False,
    ax=None,
    slider_formats=None,
    xlim="stretch",
    ylim="stretch",
    force_ipywidgets=False,
    play_buttons=None,
    controls=None,
    display_controls=True,
    **kwargs,
):
    """
    Control a plot using widgets

    interactive_plot([x], y, [fmt])

    where x/y is are either arraylike or a function that returns arrays. Any kwargs accepted by
    `matplotlib.pyplot.plot` will be passed through, other kwargs will be intrepreted as controls

    Parameters
    ----------
    x, y : array-like or scalar or function
        The horizontal / vertical coordinates of the data points.
        *x* values are optional and default to ``range(len(y))``. If both *x* and *y* are
        provided and *y* is a function then it will be called as ``y(x, **params)``. If
        *x* is a function it will be called as ``x(**params)``
    fmt : str, optional
        A format string, e.g. 'ro' for red circles. See matplotlib.pyplot.plot
        for full documentation.
        as xlim
    parametric : boolean
        If True then the function expects to have only received a value for y and that that function will
        return an array for both x and y, or will return an array with shape (N, 2)
    ax : matplotlib axis, optional
        The axis on which to plot. If none the current axis will be used.
    slider_formats : None, string, or dict
        If None a default value of decimal points will be used. Uses the new {} style formatting
    xlim : string or tuple of floats, optional
        If a tuple it will be passed to ax.set_xlim. Other options are:
        'auto': rescale the x axis for every redraw
        'stretch': only ever expand the xlims.
    ylim : string or tuple of floats, optional
        If a tuple it will be passed to ax.set_ylim. Other options are same as xlim
    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    display_controls : boolean
        Whether the controls should display on creation. Ignored if controls is specified.

    Returns
    -------
    controls

    Examples
    --------
    With numpy arrays::

        x = np.linspace(0,2*np.pi)
        tau = np.linspace(0, np.pi)
        def f(tau):
            return np.sin(x*tau)
        interactive_plot(f, tau=tau)

    with tuples::

        x = np.linspace(0,2*np.pi)
        def f(x, tau):
            return np.sin(x+tau)
        interactive_plot(x, f, tau=(0, np.pi, 1000))

    """
    kwargs, plot_kwargs = kwarg_popper(kwargs, Line2D_kwargs_list)
    x_and_y = False
    x = None
    fmt = None
    if len(args) == 0:
        # wot...
        return
    elif len(args) == 1:
        y = args[0]
    elif len(args) == 2:
        # either (y, fmt) or (x, y)
        # hard to know for sure though bc fmt can be a function
        # or maybe just requirement that fmt is a function
        if isinstance(args[1], str):
            y, fmt = args
        else:
            x_and_y = True
            x, y = args
    elif len(args) == 3:
        x_and_y = True
        x, y, fmt = args
    else:
        raise ValueError(f"You passed in {len(args)} args, but no more than 3 is supported.")

    ipympl = notebook_backend()
    ipympl or force_ipywidgets
    fig, ax = gogogo_figure(ipympl, ax=ax)
    slider_formats = create_slider_format_dict(slider_formats)
    controls, params = gogogo_controls(
        kwargs, controls, display_controls, slider_formats, play_buttons
    )

    def update(params, indices, cache):
        if x_and_y:
            x_, y_ = eval_xy(x, y, params, cache)
            # broadcast so that we can always index
            if x_.ndim == 1:
                x_ = np.broadcast_to(x_[:, None], (x_.shape[0], len(lines)))
            if y_.ndim == 1:
                y_ = np.broadcast_to(y_[:, None], (y_.shape[0], len(lines)))
            for i, line in enumerate(lines):
                line.set_data(x_[:, i], y_[:, i])
        elif parametric:
            # transpose to splat bc matplotlib considers columns of arrays to be
            # the datasets
            # I don't think it's possible to have multiple lines here
            # assert len(lines) == 1
            out = callable_else_value_no_cast(y, params, cache)
            if isinstance(out, tuple):
                pass
            elif isinstance(out, np.ndarray):
                # transpose bc set_data expects a different shape than plot
                out = np.asanyarray(out).T
            # else hope for the best lol
            lines[0].set_data(*out)
        else:
            y_ = callable_else_value(y, params, cache)
            if y_.ndim == 1:
                y_ = np.broadcast_to(y_[:, None], (y_.shape[0], len(lines)))
            for i, line in enumerate(lines):
                line.set_ydata(y_[:, i])

        cur_xlims = ax.get_xlim()
        cur_ylims = ax.get_ylim()
        ax.relim()  # this may be expensive? don't do if not necessary?
        if ylim == "auto":
            ax.autoscale_view(scalex=False)
        elif ylim == "stretch":
            new_lims = [ax.dataLim.y0, ax.dataLim.y0 + ax.dataLim.height]
            new_lims = [
                new_lims[0] if new_lims[0] < cur_ylims[0] else cur_ylims[0],
                new_lims[1] if new_lims[1] > cur_ylims[1] else cur_ylims[1],
            ]
            ax.set_ylim(new_lims)
        if xlim == "auto":
            ax.autoscale_view(scaley=False)
        elif xlim == "stretch":
            new_lims = [ax.dataLim.x0, ax.dataLim.x0 + ax.dataLim.width]
            new_lims = [
                new_lims[0] if new_lims[0] < cur_xlims[0] else cur_xlims[0],
                new_lims[1] if new_lims[1] > cur_xlims[1] else cur_xlims[1],
            ]
            ax.set_xlim(new_lims)

    controls._register_function(update, fig, params.keys())

    if x_and_y:
        x_, y_ = eval_xy(x, y, params)
        if fmt:
            lines = ax.plot(x_, y_, fmt, **plot_kwargs)
        else:
            lines = ax.plot(x_, y_, **plot_kwargs)
    else:
        y_ = callable_else_value_no_cast(y, params)
        # set up to ensure that splatting works well
        if parametric and not isinstance(y_, tuple):
            y_ = np.asanyarray(y_).T
        else:
            # make a tuple so we can splat it
            # reduces the number of if statements necessary to plot
            # parametric functions
            y_ = (y_,)

        if fmt:
            lines = ax.plot(*y_, fmt, **plot_kwargs)
        else:
            lines = ax.plot(*y_, **plot_kwargs)

    try:
        # hack in the way it feels like matplotlib should behave
        # this is a necessary change to support ODEs which is a reasonable use case for
        # this library - lesser of two evils situation. (the evil here is deviating from matplotlib)
        labels = plot_kwargs["label"]
        if (
            len(lines) > 1
            and (isinstance(labels, list) or isinstance(labels, tuple))
            and len(labels) == len(lines)
        ):
            for label, line in zip(labels, lines):
                line.set_label(label)
    except KeyError:
        pass

    if not isinstance(xlim, str):
        ax.set_xlim(xlim)
    if not isinstance(ylim, str):
        ax.set_ylim(ylim)

    # make sure the home button will work
    if hasattr(fig.canvas, "toolbar") and fig.canvas.toolbar is not None:
        fig.canvas.toolbar.push_current()
    # set current axis to be pyplot-like
    sca(ax)

    return controls


def simple_hist(arr, bins="auto", density=None, weights=None):
    heights, bins = np.histogram(arr, bins=bins, density=density, weights=weights)
    width = bins[1] - bins[0]
    new_patches = []
    for i in range(len(heights)):
        new_patches.append(Rectangle((bins[i], 0), width=width, height=heights[i]))
    xlims = (bins.min(), bins.max())
    ylims = (0, heights.max() * 1.05)

    return xlims, ylims, new_patches


def stretch(ax, xlims, ylims):
    cur_xlims = ax.get_xlim()
    cur_ylims = ax.get_ylim()
    new_lims = ylims
    new_lims = [
        new_lims[0] if new_lims[0] < cur_ylims[0] else cur_ylims[0],
        new_lims[1] if new_lims[1] > cur_ylims[1] else cur_ylims[1],
    ]
    ax.set_ylim(new_lims)
    new_lims = xlims
    new_lims = [
        new_lims[0] if new_lims[0] < cur_xlims[0] else cur_xlims[0],
        new_lims[1] if new_lims[1] > cur_xlims[1] else cur_xlims[1],
    ]
    ax.set_xlim(new_lims)


def interactive_hist(
    arr,
    density=False,
    bins="auto",
    weights=None,
    ax=None,
    slider_formats=None,
    force_ipywidgets=False,
    play_buttons=False,
    controls=None,
    display_controls=True,
    **kwargs,
):
    """
    Control the contents of a histogram using widgets.

    See https://github.com/ianhi/mpl-interactions/pull/73#issue-470638134 for a discussion
    of the limitations of this function. These limitations will be improved once
    https://github.com/matplotlib/matplotlib/pull/18275 has been merged.

    Parameters
    ----------
    arr : arraylike or function
        The array or the function that returns an array that is to be histogrammed
    density : bool, optional
        whether to plot as a probability density. Passed to `numpy.histogram`
    bins : int or sequence of scalars or str, optional
        bins argument to `numpy.histogram`
    weights : array_like, optional
        passed to `numpy.histogram`
    ax : matplotlib axis, optional
        The axis on which to plot. If none the current axis will be used.
    slider_formats : None, string, or dict
        If None a default value of decimal points will be used. Uses the new {} style formatting
    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    display_controls : boolean
        Whether the controls should display on creation. Ignored if controls is specified.

    Returns
    -------
    controls

    Examples
    --------
    With numpy arrays::

        loc = np.linspace(-5, 5, 500)
        scale = np.linspace(1, 10, 100)
        def f(loc, scale):
            return np.random.randn(1000)*scale + loc
        interactive_hist(f, loc=loc, scale=scale)

    with tuples::

        def f(loc, scale):
            return np.random.randn(1000)*scale + loc
        interactive_hist(f, loc=(-5, 5, 500), scale=(1, 10, 100))
    """

    ipympl = notebook_backend()
    fig, ax = gogogo_figure(ipympl, ax=ax)
    ipympl or force_ipywidgets
    slider_formats = create_slider_format_dict(slider_formats)
    controls, params = gogogo_controls(
        kwargs, controls, display_controls, slider_formats, play_buttons
    )
    pc = PatchCollection([])
    ax.add_collection(pc, autolim=True)

    def update(params, indices, cache):
        arr_ = callable_else_value(arr, params, cache)
        new_x, new_y, new_patches = simple_hist(arr_, density=density, bins=bins, weights=weights)
        stretch(ax, new_x, new_y)
        pc.set_paths(new_patches)
        ax.autoscale_view()

    controls._register_function(update, fig, params.keys())

    new_x, new_y, new_patches = simple_hist(
        callable_else_value(arr, params), density=density, bins=bins, weights=weights
    )
    sca(ax)
    pc.set_paths(new_patches)
    ax.set_xlim(new_x)
    ax.set_ylim(new_y)

    return controls


def interactive_scatter(
    x,
    y=None,
    s=None,
    c=None,
    cmap=None,
    vmin=None,
    vmax=None,
    alpha=None,
    marker=None,
    edgecolors=None,
    facecolors=None,
    label=None,
    parametric=False,
    ax=None,
    slider_formats=None,
    xlim="stretch",
    ylim="stretch",
    force_ipywidgets=False,
    play_buttons=False,
    controls=None,
    display_controls=True,
    **kwargs,
):
    """
    Control a scatter plot using widgets.

    Parameters
    ----------
    x, y : function or float or array-like
        shape (n, ) for array-like. Functions must return the correct shape as well. If y is None
        then parametric must be True and the function for x must return x, y
    c : array-like or list of colors or color or Callable
        Valid input to plt.scatter or a function
    s : float, array-like, function, or index controls object
        valid input to plt.scatter, or a function
    alpha : float, None, or function(s), broadcastable
        Affects all scatter points. This will compound with any alpha introduced by
        the ``c`` argument
    marker : MarkerStyle, or Callable, optional
        The marker style or a function returning marker styles.
    edgecolor[s] : callable or valid argument to scatter
        passed through to scatter.
    facecolor[s] : callable or valid argument to scatter
        Valid input to plt.scatter, or a function
    label : string
        Passed through to Matplotlib
    parametric : boolean
        If True then the function expects to have only received a value for y and that that function will
        return an array for both x and y, or will return an array with shape (N, 2)
    ax : matplotlib axis, optional
        The axis on which to plot. If none the current axis will be used.
    slider_formats : None, string, or dict
        If None a default value of decimal points will be used. Uses the new {} style formatting
    xlim : string or tuple of floats, optional
        If a tuple it will be passed to ax.set_xlim. Other options are:
        'auto': rescale the x axis for every redraw
        'stretch': only ever expand the xlims.
    ylim : string or tuple of floats, optional
        If a tuple it will be passed to ax.set_ylim. Other options are same as xlim
    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    display_controls : boolean
        Whether the controls should display on creation. Ignored if controls is specified.

    Returns
    -------
    controls
    """

    if isinstance(xlim, str):
        stretch_x = xlim == "stretch"
    else:
        stretch_x = False

    if isinstance(ylim, str) and ylim.lower() == "stretch":
        stretch_y = True
    else:
        stretch_y = False

    # yanked from https://github.com/matplotlib/matplotlib/blob/bcc1ce8461f5b6e874baaaa02ef776d0243a4abe/lib/matplotlib/axes/_axes.py#L4271-L4273
    facecolors = kwargs.pop("facecolor", facecolors)
    edgecolors = kwargs.pop("edgecolor", edgecolors)

    kwargs, collection_kwargs = kwarg_popper(kwargs, collection_kwargs_list)

    ipympl = notebook_backend()
    fig, ax = gogogo_figure(ipympl, ax)
    ipympl or force_ipywidgets
    slider_formats = create_slider_format_dict(slider_formats)

    extra_ctrls = []
    funcs, extra_ctrls, param_excluder = prep_scalars(kwargs, s=s, alpha=alpha, marker=marker)
    s = funcs["s"]
    alpha = funcs["alpha"]
    marker = funcs["marker"]

    controls, params = gogogo_controls(
        kwargs, controls, display_controls, slider_formats, play_buttons, extra_ctrls
    )

    def update(params, indices, cache):
        if parametric:
            out = callable_else_value_no_cast(x, param_excluder(params))
            if not isinstance(out, tuple):
                out = np.asanyarray(out).T
            x_, y_ = out
        else:
            x_, y_ = eval_xy(x, y, param_excluder(params), cache)
        scatter.set_offsets(np.column_stack([x_, y_]))
        c_ = check_callable_xy(c, x_, y_, param_excluder(params), cache)
        s_ = check_callable_xy(s, x_, y_, param_excluder(params, "s"), cache)
        ec_ = check_callable_xy(edgecolors, x_, y_, param_excluder(params), cache)
        fc_ = check_callable_xy(facecolors, x_, y_, param_excluder(params), cache)
        a_ = check_callable_alpha(alpha, param_excluder(params, "alpha"), cache)
        marker_ = callable_else_value_no_cast(marker, param_excluder(params), cache)

        if marker_ is not None:
            if not isinstance(marker_, mmarkers.MarkerStyle):
                marker_ = mmarkers.MarkerStyle(marker_)
            path = marker_.get_path().transformed(marker_.get_transform())
            scatter.set_paths((path,))

        if c_ is not None:
            try:
                c_ = to_rgba_array(c_)
            except ValueError:
                try:
                    c_ = scatter.cmap(c_)
                except TypeError:
                    raise ValueError(
                        "If c is a function it must return either an RGB(A) array"
                        "or a 1D array of valid color names or values to be colormapped"
                    )
            scatter.set_facecolor(c_)
        if ec_ is not None:
            scatter.set_edgecolor(ec_)
        if fc_ is not None:
            scatter.set_facecolor(c_)
        if s_ is not None:
            if isinstance(s_, Number):
                s_ = np.broadcast_to(s_, (len(x_),))
            scatter.set_sizes(s_)
        if a_ is not None:
            scatter.set_alpha(a_)

        update_datalim_from_bbox(
            ax, scatter.get_datalim(ax.transData), stretch_x=stretch_x, stretch_y=stretch_y
        )
        ax.autoscale_view()

    controls._register_function(update, fig, params.keys())

    def check_callable_xy(arg, x, y, params, cache):
        if isinstance(arg, Callable):
            if arg not in cache:
                cache[arg] = arg(x, y, **params)
            return cache[arg]
        else:
            return arg

    def check_callable_alpha(alpha_, params, cache):
        if isinstance(alpha_, Callable):
            if alpha_ not in cache:
                cache[alpha_] = alpha_(**param_excluder(params, "alpha"))
            return cache[alpha_]
        else:
            return alpha_

    p = param_excluder(params)
    if parametric:
        out = callable_else_value_no_cast(x, p)
        if not isinstance(out, tuple):
            out = np.asanyarray(out).T
        x_, y_ = out
    else:
        x_, y_ = eval_xy(x, y, p)
    c_ = check_callable_xy(c, x_, y_, p, {})
    s_ = check_callable_xy(s, x_, y_, param_excluder(params, "s"), {})
    ec_ = check_callable_xy(edgecolors, x_, y_, p, {})
    fc_ = check_callable_xy(facecolors, x_, y_, p, {})
    a_ = check_callable_alpha(alpha, params, {})
    marker_ = callable_else_value_no_cast(marker, p, {})
    scatter = ax.scatter(
        x_,
        y_,
        c=c_,
        s=s_,
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
        marker=marker_,
        alpha=a_,
        edgecolors=ec_,
        facecolors=fc_,
        label=label,
        **collection_kwargs,
    )
    # this is necessary to make calls to plt.colorbar behave as expected
    sca(ax)
    ax._sci(scatter)

    return controls


# portions of this docstring were copied directly from the docsting
# of `matplotlib.pyplot.imshow`
def interactive_imshow(
    X,
    cmap=None,
    norm=None,
    aspect=None,
    interpolation=None,
    alpha=None,
    vmin=None,
    vmax=None,
    vmin_vmax=None,
    origin=None,
    extent=None,
    autoscale_cmap=True,
    filternorm=True,
    filterrad=4.0,
    resample=None,
    url=None,
    ax=None,
    slider_formats=None,
    force_ipywidgets=False,
    play_buttons=False,
    controls=None,
    display_controls=True,
    **kwargs,
):
    """
    Control an image using widgets.

    Parameters
    ----------
    X : function or image like
        If a function it must return an image-like object. See matplotlib.pyplot.imshow for the
        full set of valid options.
    cmap : str or `~matplotlib.colors.Colormap`
        The Colormap instance or registered colormap name used to map
        scalar data to colors. This parameter is ignored for RGB(A) data.
        forwarded to matplotlib
    norm : `~matplotlib.colors.Normalize`, optional
        The `~matplotlib.colors.Normalize` instance used to scale scalar data to the [0, 1]
        range before mapping to colors using *cmap*. By default, a linear
        scaling mapping the lowest value to 0 and the highest to 1 is used.
        This parameter is ignored for RGB(A) data.
        forwarded to matplotlib
    autoscale_cmap : bool
        If True rescale the colormap for every function update. Will not update
        if vmin and vmax are provided or if the returned image is RGB(A) like.
        forwarded to matplotlib
    aspect : {'equal', 'auto'} or float
        forwarded to matplotlib
    interpolation : str
        forwarded to matplotlib
    alpha : float, callable, shorthand for slider or indexed controls
        The alpha value of the image. Can accept a float for a fixed value,
        or any slider shorthand to control with a slider, or an indexed controls
        object to use an existing slider, or an arbitrary function of the other
        parameters.
    ax : matplotlib axis, optional
        The axis on which to plot. If none the current axis will be used.
    slider_formats : None, string, or dict
        If None a default value of decimal points will be used. Uses the new {} style formatting
    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    display_controls : boolean
        Whether the controls should display on creation. Ignored if controls is specified.

    Returns
    -------
    controls
    """
    ipympl = notebook_backend()
    fig, ax = gogogo_figure(ipympl, ax)
    ipympl or force_ipywidgets
    slider_formats = create_slider_format_dict(slider_formats)
    kwargs, imshow_kwargs = kwarg_popper(kwargs, imshow_kwargs_list)

    funcs, extra_ctrls, param_excluder = prep_scalars(kwargs, vmin=vmin, vmax=vmax, alpha=alpha)
    vmin = funcs["vmin"]
    vmax = funcs["vmax"]
    alpha = funcs["alpha"]

    if vmin_vmax is not None:
        if isinstance(vmin_vmax, tuple) and not isinstance(vmin_vmax[0], str):
            vmin_vmax = ("r", *vmin_vmax)
        kwargs["vmin_vmax"] = vmin_vmax

    controls, params = gogogo_controls(
        kwargs, controls, display_controls, slider_formats, play_buttons, extra_ctrls
    )
    if vmin_vmax is not None:
        params.pop("vmin_vmax")
        params["vmin"] = controls.params["vmin"]
        params["vmax"] = controls.params["vmax"]

        def vmin(**kwargs):
            return kwargs["vmin"]

        def vmax(**kwargs):
            return kwargs["vmax"]

    def update(params, indices, cache):
        if isinstance(X, Callable):
            # ignore anything that we added directly to kwargs in prep_scalar
            # if we don't do this then we might pass the user a kwarg their function
            # didn't expect and things may break
            # check this here to avoid setting the data if we don't need to
            # use the callable_else_value fxn to make use of easy caching
            new_data = callable_else_value(X, param_excluder(params), cache)
            im.set_data(new_data)
            if autoscale_cmap and (new_data.ndim != 3) and vmin is None and vmax is None:
                im.norm.autoscale(new_data)
        # caching for these?
        if isinstance(vmin, Callable):
            im.norm.vmin = callable_else_value(vmin, param_excluder(params, "vmin"), cache)
        if isinstance(vmax, Callable):
            im.norm.vmax = callable_else_value(vmax, param_excluder(params, "vmax"), cache)
        # Don't use callable_else_value to avoid unnecessary updates
        # Seems as though set_alpha doesn't short circuit if the value
        # hasn't been changed
        if isinstance(alpha, Callable):
            im.set_alpha(callable_else_value_no_cast(alpha, param_excluder(params, "alpha"), cache))

    controls._register_function(update, fig, params.keys())

    # make it once here so we can use the dims in update
    # see explanation for excluded_params in the update function
    new_data = callable_else_value(X, param_excluder(params))
    sca(ax)
    im = ax.imshow(
        new_data,
        cmap=cmap,
        norm=norm,
        aspect=aspect,
        interpolation=interpolation,
        alpha=callable_else_value_no_cast(alpha, param_excluder(params, "alpha")),
        vmin=callable_else_value_no_cast(vmin, param_excluder(params, "vmin")),
        vmax=callable_else_value_no_cast(vmax, param_excluder(params, "vmax")),
        origin=origin,
        extent=extent,
        filternorm=filternorm,
        filterrad=filterrad,
        resample=resample,
        url=url,
        **imshow_kwargs,
    )

    # i know it's bad news to use private methods :(
    # but idk how else to accomplish being a psuedo-pyplot
    ax._sci(im)
    return controls


def interactive_axhline(
    y=0,
    xmin=0,
    xmax=1,
    ax=None,
    slider_formats=None,
    force_ipywidgets=False,
    play_buttons=False,
    controls=None,
    display_controls=True,
    **kwargs,
):
    """
    Control an horizontal line using widgets.

    Parameters
    ----------
    y : float or function
        y position in data coordinates of the horizontal line.
    xmin : float or function
        Should be between 0 and 1, 0 being the far left of the plot, 1 the
        far right of the plot.
    xmax : float or function
        Should be between 0 and 1, 0 being the far left of the plot, 1 the
        far right of the plot.
    ax : matplotlib axis, optional
        If None a new figure and axis will be created
    slider_formats : None, string, or dict
        If None a default value of decimal points will be used. Uses the new {} style formatting
    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    display_controls : boolean
        Whether the controls should display on creation. Ignored if controls is specified.
    **kwargs
        Kwargs will be used to create control widgets. Except kwargs that are valid for Line2D are
        extracted and passed through to the creation of the line.

    Returns
    -------
    controls
    """
    ipympl = notebook_backend()
    fig, ax = gogogo_figure(ipympl, ax)
    ipympl or force_ipywidgets
    slider_formats = create_slider_format_dict(slider_formats)

    kwargs, line_kwargs = kwarg_popper(kwargs, Line2D_kwargs_list)
    line_kwargs.pop("transform", None)  # transform is not a valid kwarg for ax{v,h}line

    extra_ctrls = []
    funcs, extra_ctrls, param_excluder = prep_scalars(kwargs, y=y, xmin=xmin, xmax=xmax)
    y = funcs["y"]
    xmin = funcs["xmin"]
    xmax = funcs["xmax"]
    controls, params = gogogo_controls(
        kwargs, controls, display_controls, slider_formats, play_buttons, extra_ctrls
    )

    def update(params, indices, cache):
        y_ = callable_else_value(y, param_excluder(params, "y"), cache).item()
        line.set_ydata([y_, y_])
        xmin_ = callable_else_value(xmin, param_excluder(params, "xmin"), cache).item()
        xmax_ = callable_else_value(xmax, param_excluder(params, "xmax"), cache).item()
        line.set_xdata([xmin_, xmax_])
        # TODO consider updating just the ydatalim here

    controls._register_function(update, fig, params)
    sca(ax)
    line = ax.axhline(
        callable_else_value(y, param_excluder(params, "y")).item(),
        callable_else_value(xmin, param_excluder(params, "xmin")).item(),
        callable_else_value(xmax, param_excluder(params, "xmax")).item(),
        **line_kwargs,
    )
    return controls


def interactive_axvline(
    x=0,
    ymin=0,
    ymax=1,
    ax=None,
    slider_formats=None,
    force_ipywidgets=False,
    play_buttons=False,
    controls=None,
    display_controls=True,
    **kwargs,
):
    """
    Control a vertical line using widgets.

    Parameters
    ----------
    x : float or function
        x position in data coordinates of the horizontal line.
    ymin : float or function
        Should be between 0 and 1, 0 being the bottom of the plot, 1 the
        far top of the plot
    ymax : float or function
        Should be between 0 and 1, 0 being the top of the plot, 1 the
        top of the plot.
    ax : matplotlib axis, optional
        If None a new figure and axis will be created
    slider_formats : None, string, or dict
        If None a default value of decimal points will be used. Uses the new {} style formatting
    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    display_controls : boolean
        Whether the controls should display on creation. Ignored if controls is specified.
    **kwargs
        Kwargs will be used to create control widgets. Except kwargs that are valid for Line2D are
        extracted and passed through to the creation of the line.

    Returns
    -------
    controls
    """
    ipympl = notebook_backend()
    fig, ax = gogogo_figure(ipympl, ax)
    ipympl or force_ipywidgets
    slider_formats = create_slider_format_dict(slider_formats)

    kwargs, line_kwargs = kwarg_popper(kwargs, Line2D_kwargs_list)
    line_kwargs.pop("transform", None)  # transform is not a valid kwarg for ax{v,h}line

    extra_ctrls = []
    funcs, extra_ctrls, param_excluder = prep_scalars(kwargs, x=x, ymin=ymin, ymax=ymax)
    x = funcs["x"]
    ymin = funcs["ymin"]
    ymax = funcs["ymax"]

    controls, params = gogogo_controls(
        kwargs, controls, display_controls, slider_formats, play_buttons, extra_ctrls
    )

    def update(params, indices, cache):
        x_ = callable_else_value(x, param_excluder(params, "x"), cache).item()
        line.set_xdata([x_, x_])
        ymin_ = callable_else_value(ymin, param_excluder(params, "ymin"), cache).item()
        ymax_ = callable_else_value(ymax, param_excluder(params, "ymax"), cache).item()
        line.set_ydata([ymin_, ymax_])
        # TODO consider updating just the ydatalim here

    controls._register_function(update, fig, params)
    sca(ax)
    line = ax.axvline(
        callable_else_value(x, param_excluder(params, "x")).item(),
        callable_else_value(ymin, param_excluder(params, "ymin")).item(),
        callable_else_value(ymax, param_excluder(params, "ymax")).item(),
        **line_kwargs,
    )
    return controls


def interactive_title(
    title,
    controls=None,
    ax=None,
    *,
    fontdict=None,
    loc=None,
    y=None,
    pad=None,
    slider_formats=None,
    display_controls=True,
    play_buttons=False,
    force_ipywidgets=False,
    **kwargs,
):
    """
    Set an title that will update interactively. kwargs for `matplotlib.text.Text` will be passed through,
    other kwargs will be used to create interactive controls.

    Parameters
    ----------
    title : str or function
        The title text. Can include {} style formatting. e.g. 'The voltage is {volts:.2f} mV'
    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    ax : `matplotlib.axes.Axes`, optional
        The axis on which to plot. If none the current axis will be used.
    loc : {'center', 'left', 'right'}, default: `axes.titlelocation <matplotlib.rcParams>`
        Which title to set.
    y : float, default: `axes.titley <matplotlib.rcParams>`
        Vertical axes loation for the title (1.0 is the top).  If
        None (the default), y is determined automatically to avoid
        decorators on the axes.
    pad : float, default: `axes.titlepad <matplotlib.rcParams>`
        The offset of the title from the top of the axes, in points.
    slider_formats : None, string, or dict
        If None a default value of decimal points will be used. Uses {} style formatting
    display_controls : boolean
        Whether the controls should display on creation. Ignored if controls is specified.
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets

    Returns
    -------
    controls
    """
    ipympl = notebook_backend()
    fig, ax = gogogo_figure(ipympl, ax)
    ipympl or force_ipywidgets
    slider_formats = create_slider_format_dict(slider_formats)

    kwargs, text_kwargs = kwarg_popper(kwargs, Text_kwargs_list)

    controls, params = gogogo_controls(
        kwargs, controls, display_controls, slider_formats, play_buttons
    )

    def update(params, indices, cache):
        ax.set_title(
            callable_else_value_no_cast(title, params, cache).format(**params),
            fontdict=fontdict,
            loc=loc,
            pad=pad,
            y=y,
            **text_kwargs,
        )

    controls._register_function(update, fig, params)
    ax.set_title(
        callable_else_value_no_cast(title, params, None).format(**params),
        fontdict=fontdict,
        loc=loc,
        pad=pad,
        y=y,
        **text_kwargs,
    )
    return controls


def interactive_xlabel(
    xlabel,
    controls=None,
    ax=None,
    *,
    fontdict=None,
    labelpad=None,
    loc=None,
    slider_formats=None,
    display_controls=True,
    play_buttons=False,
    force_ipywidgets=False,
    **kwargs,
):
    """
    Set an xlabel that will update interactively. kwargs for `matplotlib.text.Text` will be passed through,
    other kwargs will be used to create interactive controls.

    Parameters
    ----------
    xlabel : str or function
        The label text. Can include {} style formatting. e.g. 'The voltage is {volts:.2f} mV'
    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    ax : matplotlib axis, optional
        The axis on which to plot. If none the current axis will be used.
    labelpad : float, default: None
        Spacing in points from the axes bounding box including ticks
        and tick labels.
    loc : {'bottom', 'center', 'top'}, default: `yaxis.labellocation <matplotlib.rcParams>`
        The label position. This is a high-level alternative for passing
        parameters *y* and *horizontalalignment*.
    slider_formats : None, string, or dict
        If None a default value of decimal points will be used. Uses {} style formatting
    display_controls : boolean
        Whether the controls should display on creation. Ignored if controls is specified.
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets

    Returns
    -------
    controls
    """
    ipympl = notebook_backend()
    fig, ax = gogogo_figure(ipympl, ax)
    ipympl or force_ipywidgets
    slider_formats = create_slider_format_dict(slider_formats)

    kwargs, text_kwargs = kwarg_popper(kwargs, Text_kwargs_list)
    controls, params = gogogo_controls(
        kwargs, controls, display_controls, slider_formats, play_buttons
    )

    def update(params, indices, cache):
        ax.set_xlabel(
            callable_else_value_no_cast(xlabel, params, cache).format(**params),
            fontdict=fontdict,
            labelpad=labelpad,
            loc=loc,
            **text_kwargs,
        )

    controls._register_function(update, fig, params)
    ax.set_xlabel(
        callable_else_value_no_cast(xlabel, params, None).format(**params),
        fontdict=fontdict,
        labelpad=labelpad,
        loc=loc,
        **text_kwargs,
    )
    return controls


def interactive_ylabel(
    ylabel,
    controls=None,
    ax=None,
    *,
    fontdict=None,
    labelpad=None,
    loc=None,
    slider_formats=None,
    display_controls=True,
    play_buttons=False,
    force_ipywidgets=False,
    **kwargs,
):
    """
    Set a ylabel that will update interactively. kwargs for `matplotlib.text.Text` will be passed through,
    other kwargs will be used to create interactive controls.

    Parameters
    ----------
    ylabel : str or function
        The label text. Can include {} style formatting. e.g. 'The voltage is {volts:.2f}'
    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    ax : matplotlib axis, optional
        The axis on which to plot. If none the current axis will be used.
    labelpad : float, default: None
        Spacing in points from the axes bounding box including ticks
        and tick labels.
    loc : {'bottom', 'center', 'top'}, default: `yaxis.labellocation <matplotlib.rcParams>`
        The label position. This is a high-level alternative for passing
        parameters *y* and *horizontalalignment*.
    slider_formats : None, string, or dict
        If None a default value of decimal points will be used. Uses {} style formatting
    display_controls : boolean
        Whether the controls should display on creation. Ignored if controls is specified.
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets

    Returns
    -------
    controls
    """
    ipympl = notebook_backend()
    fig, ax = gogogo_figure(ipympl, ax)
    ipympl or force_ipywidgets
    slider_formats = create_slider_format_dict(slider_formats)

    kwargs, text_kwargs = kwarg_popper(kwargs, Text_kwargs_list)
    controls, params = gogogo_controls(
        kwargs, controls, display_controls, slider_formats, play_buttons
    )

    def update(params, indices, cache):
        ax.set_ylabel(
            callable_else_value_no_cast(ylabel, params, cache).format(**params),
            fontdict=fontdict,
            labelpad=labelpad,
            loc=loc,
            **text_kwargs,
        )

    controls._register_function(update, fig, params)
    ax.set_ylabel(
        callable_else_value_no_cast(ylabel, params, None).format(**params),
        fontdict=fontdict,
        labelpad=labelpad,
        loc=loc,
        **text_kwargs,
    )
    return controls


def interactive_text(
    x,
    y,
    s,
    fontdict=None,
    controls=None,
    ax=None,
    *,
    slider_formats=None,
    display_controls=True,
    play_buttons=False,
    force_ipywidgets=False,
    **kwargs,
):
    """
    Create a text object that will update interactively.
    kwargs for `matplotlib.text.Text` will be passed through, other kwargs will be used to create interactive controls.

    .. note::

        fontdict properties are currently static - see https://github.com/ianhi/mpl-interactions/issues/247


    Parameters
    ----------
    x, y : float or function
        The text position.
    s : str or function
        The text. Can either be static text, a function returning a string or
        can include {} style formatting. e.g. 'The voltage is {volts:.2f}'
    fontdict : dict[str]
        Passed through to the Text object. Currently not dynamically updateable. See
        https://github.com/ianhi/mpl-interactions/issues/247
    controls : mpl_interactions.controller.Controls
        An existing controls object if you want to tie multiple plot elements to the same set of
        controls
    ax : matplotlib axis, optional
        The axis on which to plot. If none the current axis will be used.
    play_buttons : bool or str or dict, optional
        Whether to attach an ipywidgets.Play widget to any sliders that get created.
        If a boolean it will apply to all kwargs, if a dictionary you choose which sliders you
        want to attach play buttons too.

        - None: no sliders
        - True: sliders on the lft
        - False: no sliders
        - 'left': sliders on the left
        - 'right': sliders on the right

    force_ipywidgets : boolean
        If True ipywidgets will always be used, even if not using the ipympl backend.
        If False the function will try to detect if it is ok to use ipywidgets
        If ipywidgets are not used the function will fall back on matplotlib widgets

    Returns
    -------
    controls
    """
    ipympl = notebook_backend()
    fig, ax = gogogo_figure(ipympl, ax)
    ipympl or force_ipywidgets
    slider_formats = create_slider_format_dict(slider_formats)

    kwargs, text_kwargs = kwarg_popper(kwargs, Text_kwargs_list)
    funcs, extra_ctrls, param_excluder = prep_scalars(kwargs, x=x, y=y)
    x = funcs["x"]
    y = funcs["y"]
    controls, params = gogogo_controls(
        kwargs, controls, display_controls, slider_formats, play_buttons, extra_ctrls
    )

    def update(params, indices, cache):
        x_, y_ = eval_xy(x, y, param_excluder(params, ["x", "y"]), cache)
        text.set_x(x_)
        text.set_y(y_)
        text.set_text(callable_else_value_no_cast(s, params, cache).format(**params))

    controls._register_function(update, fig, params)
    x_, y_ = eval_xy(x, y, param_excluder(params, ["x", "y"]))
    text = ax.text(
        x_,
        y_,
        callable_else_value_no_cast(s, params).format(**params),
        fontdict=fontdict,
        **text_kwargs,
    )
    return controls



# cribbing from matplotlib's approach
# by default DeprecationWarnings are ignored. Define our own class to make sure they are seen

__all__ = ["mpl_interactions_DeprecationWarning"]


class mpl_interactions_DeprecationWarning(DeprecationWarning):
    """A class for issuing deprecation warnings for mpl-interactions users."""



"""
based on https://matplotlib.org/3.1.0/gallery/user_interfaces/embedding_in_tk_sgskip.html
special thanks to @tacaswell for helping on the matplotlib discourse:
https://discourse.matplotlib.org/t/mpl-iteractions-with-gui-outside-of-jupyter-notebooks/21523/6
"""

import tkinter

import numpy as np
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk

# Implement the default Matplotlib key bindings.
from matplotlib.figure import Figure

from mpl_interactions import heatmap_slicer

root = tkinter.Tk()
root.wm_title("Embedding heatmap_slicer Tk gui")

fig = Figure(figsize=(5, 4), dpi=100)


canvas = FigureCanvasTkAgg(fig, master=root)  # A tk.DrawingArea.
canvas.draw()
canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)

toolbar = NavigationToolbar2Tk(canvas, root)
toolbar.update()
canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)


x = np.linspace(0, np.pi, 100)
y = np.linspace(0, 10, 200)
X, Y = np.meshgrid(x, y)
data1 = np.sin(X) + np.exp(np.cos(Y))
data2 = np.cos(X) + np.exp(np.sin(Y))
_, axes = heatmap_slicer(
    x,
    y,
    (data1, data2),
    slices="both",
    heatmap_names=("dataset 1", "dataset 2"),
    labels=("Some wild X variable", "Y axis"),
    interaction_type="move",
    fig=fig,
)


def _quit():
    root.quit()  # stops mainloop
    root.destroy()  # this is necessary on Windows to prevent
    # Fatal Python Error: PyEval_RestoreThread: NULL tstate


button = tkinter.Button(master=root, text="Quit", command=_quit)
button.pack(side=tkinter.BOTTOM)

tkinter.mainloop()



"""
=====================================================
Matplotlib Sliders without a separate Controls Figure
=====================================================

Demonstration of how to provide a matplotlib slider to prevent
the creation of a separate controls figure.
"""
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Slider

from mpl_interactions import ipyplot as iplt

fig, ax = plt.subplots()
plt.subplots_adjust(bottom=0.25)
x = np.linspace(0, 2 * np.pi, 200)


def f(x, freq):
    return np.sin(x * freq)


axfreq = plt.axes([0.25, 0.1, 0.65, 0.03])
slider = Slider(axfreq, label="freq", valmin=0.05, valmax=10)
controls = iplt.plot(x, f, freq=slider, ax=ax)
plt.show()



try:
    from IPython.display import display as ipy_display
    from ipywidgets import widgets

    _not_ipython = False
except ImportError:
    _not_ipython = True
from collections import defaultdict
from collections.abc import Iterable
from functools import partial

from matplotlib.animation import FuncAnimation
from matplotlib.widgets import AxesWidget
from matplotlib.widgets import Slider as mSlider

from .helpers import (
    create_mpl_controls_fig,
    create_slider_format_dict,
    kwarg_to_ipywidget,
    kwarg_to_mpl_widget,
    notebook_backend,
    process_mpl_widget,
)


class Controls:
    def __init__(
        self,
        slider_formats=None,
        play_buttons=False,
        play_button_pos="right",
        use_ipywidgets=None,
        use_cache=True,
        **kwargs,
    ):
        # it might make sense to also accept kwargs as a straight up arg
        # to allow for passing the dictionary, but then it would need a different name
        # and we'd have to combine dicitonarys which looks like a hassle

        if use_ipywidgets is None:
            # if this ends up being true we are garunteed
            self.use_ipywidgets = notebook_backend()
        else:
            self.use_ipywidgets = use_ipywidgets
        if self.use_ipywidgets:
            if _not_ipython:
                raise ValueError(
                    "You need to be in an Environment with IPython.display available to use ipywidgets"
                )
            self.vbox = widgets.VBox([])
        else:
            self.control_figures = []
            """Storage for figures made of matplotlib sliders."""

        self.use_cache = use_cache
        self.kwargs = kwargs
        self.slider_format_strings = create_slider_format_dict(slider_formats)
        self.controls = {}
        self.params = {}
        """Parameters in the controller, see :doc:`/examples/custom-callbacks`."""
        self.figs = defaultdict(list)  # maybe should only store weakrefs?
        self.indices = defaultdict(lambda: 0)
        self._update_funcs = defaultdict(list)
        self._user_callbacks = defaultdict(list)
        self.add_kwargs(kwargs, slider_formats, play_buttons)

    def add_kwargs(self, kwargs, slider_formats=None, play_buttons=None, allow_duplicates=False):
        """
        If you pass a redundant kwarg it will just be overwritten
        maybe should only raise a warning rather than an error?

        need to implement matplotlib widgets
        also a big question is how to dynamically update the display of matplotlib widgets.
        """
        if isinstance(play_buttons, bool) or isinstance(play_buttons, str) or play_buttons is None:
            _play_buttons = defaultdict(lambda: play_buttons)
        elif isinstance(play_buttons, defaultdict):
            _play_buttons = play_buttons
        elif isinstance(play_buttons, dict):
            _play_buttons = defaultdict(lambda: False, play_buttons)
        elif isinstance(play_buttons, Iterable) and all([isinstance(p, str) for p in play_buttons]):
            _play_buttons = defaultdict(
                lambda: False, dict(zip(play_buttons, [True] * len(play_buttons)))
            )
        else:
            _play_buttons = play_buttons
        if slider_formats is not None:
            slider_formats = create_slider_format_dict(slider_formats)
            for k, v in slider_formats.items():
                self.slider_format_strings[k] = v
        if self.use_ipywidgets:
            for k, v in kwargs.items():
                if k in self.params:
                    if allow_duplicates:
                        continue
                    else:
                        raise ValueError("can't overwrite an existing param in the controller")
                if isinstance(v, AxesWidget):
                    self.params[k], self.controls[k], _ = process_mpl_widget(
                        v, partial(self.slider_updated, key=k)
                    )
                else:
                    self.params[k], control = kwarg_to_ipywidget(
                        k,
                        v,
                        partial(self.slider_updated, key=k),
                        self.slider_format_strings[k],
                        play_button=_play_buttons[k],
                    )
                    if control:
                        self.controls[k] = control
                        self.vbox.children = list(self.vbox.children) + [control]
                if k == "vmin_vmax":
                    self.params["vmin"] = self.params["vmin_vmax"][0]
                    self.params["vmax"] = self.params["vmin_vmax"][1]
        else:
            if len(kwargs) > 0:
                mpl_layout = create_mpl_controls_fig(kwargs)
                self.control_figures.append(mpl_layout[0])
                widget_y = 0.05
                for k, v in kwargs.items():
                    if k in self.params:
                        if allow_duplicates:
                            continue
                        else:
                            raise ValueError("Can't overwrite an existing param in the controller")
                    self.params[k], control, cb, widget_y = kwarg_to_mpl_widget(
                        mpl_layout[0],
                        mpl_layout[1:],
                        widget_y,
                        k,
                        v,
                        partial(self.slider_updated, key=k),
                        self.slider_format_strings[k],
                    )
                    if control:
                        self.controls[k] = control
                    if k == "vmin_vmax":
                        self.params["vmin"] = self.params["vmin_vmax"][0]
                        self.params["vmax"] = self.params["vmin_vmax"][1]

    def _slider_updated(self, change, key, values):
        """
        gotta also give the indices in order to support hyperslicer without horrifying contortions
        """
        if values is None:
            self.params[key] = change["new"]
        else:
            c = change["new"]
            # int casting due to a bug in numpy < 1.19
            # see https://github.com/ianhi/mpl-interactions/pull/155
            if isinstance(c, Iterable):
                # range sliders return two indices
                # check for iterable as mpl and ipywidgets sliders don't both use
                # tuples - https://github.com/ianhi/mpl-interactions/issues/195
                self.params[key] = values[[int(c) for c in change["new"]]]
                if key == "vmin_vmax":
                    self.params["vmin"] = self.params[key][0]
                    self.params["vmax"] = self.params[key][1]
            else:
                self.params[key] = values[int(change["new"])]

        self.indices[key] = change["new"]
        if self.use_cache:
            cache = {}
        else:
            cache = None

        for f, params in self._update_funcs[key]:
            ps = {}
            idxs = {}
            for k in params:
                ps[k] = self.params[k]
                idxs[k] = self.indices[k]
            f(params=ps, indices=idxs, cache=cache)
        for f, params in self._user_callbacks[key]:
            f(**{key: self.params[key] for key in params})

        for f in self.figs[key]:
            f.canvas.draw_idle()

    def slider_updated(self, change, key, values):
        """
        thin wrapper to enable splitting of special cased range sliders.
        e.g. of ``vmin_vmax`` -> ``vmin`` and ``vmax``. In the future maybe
        generalize this to any range slider with an underscore in the name?
        """
        self._slider_updated(change, key, values)
        if key == "vmin_vmax":
            self._slider_updated({"new": change["new"][0]}, "vmin", values)
            self._slider_updated({"new": change["new"][1]}, "vmax", values)

    def register_callback(self, callback, params=None, eager=False):
        """
        Register a callback to be called anytime one of the specified params changes.

        Parameters
        ----------
        callback : callable
            A function called. Should accept all of the parameters specified by *params*
            as a kwargs.
        params : str, list of str, or None
            The params to be passed to the callback. If *None* then all params
            currently registered with this controls object will be used.
        eager : bool, default: False
            If True, call the callback immediately upon registration
        """
        if isinstance(params, str):
            params = [params]
        if eager:
            if params is None:
                callback(**self.params)
            else:
                callback(**{key: self.params[key] for key in params})
        self._register_function(callback, fig=None, params=params)

    def _register_function(self, f, fig=None, params=None):
        """
        if params is None use the entire current set of params
        """
        if params is None:
            params = self.params.keys()
        # listify to ensure it's not a reference to dicts keys
        # bc that's mutable
        params = list(params)
        for p in params:
            if fig is None:
                self._user_callbacks[p].append((f, params))
            else:
                self._update_funcs[p].append((f, params))
                if fig not in self.figs[p] and fig is not None:
                    self.figs[p].append(fig)  # maybe should use a weakref?
                    # also should probably register a close_event callback to remove
                    # the figure

    def save_animation(
        self, filename, fig, param, interval=20, func_anim_kwargs={}, N_frames=None, **kwargs
    ):
        """
        Save an animation over one of the parameters controlled by this `Controls` object.

        Parameters
        ----------
        filename : str
        fig : figure
        param : str
            the name of the kwarg to use to animate
        interval : int, default: 2o
            interval between frames in ms
        func_anim_kwargs : dict
            kwargs to pass the creation of the underlying FuncAnimation
        N_frames : int
            Only used if the param is a matplotlib slider that was created without a
            valstep argument. This will only be relevant if you passed your own matplotlib
            slider as a kwarg when plotting. If needed but not given it will default to
            a value of 200.
        **kwargs
            Passed through to anim.save

        Returns
        -------
        anim : matplotlib.animation.FuncAniation
        """
        slider = self.controls[param]
        ipywidgets_slider = False
        if "Box" in str(slider.__class__):
            for obj in slider.children:
                if "Slider" in str(obj.__class__):
                    slider = obj

        if isinstance(slider, mSlider):
            min_ = slider.valmin
            max_ = slider.valmax
            if slider.valstep is None:
                n_steps = N_frames if N_frames else 200
                step = (max_ - min_) / n_steps
            else:
                step = slider.valstep
        elif "Slider" in str(slider.__class__):
            ipywidgets_slider = True
            min_ = slider.min
            max_ = slider.max
            step = slider.step
        else:
            raise NotImplementedError(
                "Cannot save animation for slider of type {slider.__class__.__name__}"
            )

        N = int((max_ - min_) / step)

        def f(i):
            val = min_ + step * i
            if ipywidgets_slider:
                slider.value = val
            else:
                slider.set_val(val)
            return []

        repeat = func_anim_kwargs.pop("repeat", False)
        anim = FuncAnimation(fig, f, frames=N, interval=interval, repeat=repeat, **func_anim_kwargs)
        # draw then stop necessary to prevent an extra loop after finished saving
        # see https://discourse.matplotlib.org/t/how-to-prevent-funcanimation-looping-a-single-time-after-save/21680/2
        fig.canvas.draw()
        anim.event_source.stop()
        anim.save(filename, **kwargs)
        return anim

    def display(self):
        """
        Display the display the ipywidgets controls or show the control figures
        """
        if self.use_ipywidgets:
            ipy_display(self.vbox)
        else:
            for fig in self.control_figures:
                if fig is not None:
                    fig.show()

    def show(self):
        """
        Show the control figures or display the ipywidgets controls
        """
        self.display()

    def _ipython_display_(self):
        ipy_display(self.vbox)

    def __getitem__(self, key):
        """
        hack to allow calls like
        interactive_plot(...beta=(0,1), controls = controls["tau"])
        also allows [None] to grab None of the current params
        to imply that we only want tau from the existing set of commands

        I think ideally this would give another controls object with just the given
        params that has this one as a parent - I think that that is most consistent with
        the idea of indexing (e.g. indexing a numpy array gives you a numpy array).
        But it's not clear how to implement that with all the sliders and such that get
        created. So for now do a sort of half-measure by returing the controls_proxy object.
        """

        # make sure keys is a list
        # bc in gogogo_controls it may get added to another list
        keys = key
        if isinstance(key, str):
            keys = [key]
        if keys is not None:
            for k in keys:
                if k not in self.params:
                    raise IndexError(f"{k} is not a param in this Controls object.")

        return _controls_proxy(self, context=False, keys=keys)

    def __enter__(self):
        self._context = _controls_proxy(self, context=True, keys=list(self.params.keys()))
        return self._context

    def __exit__(self, exc_type, exc_value, traceback):
        self._context._stack.remove(self._context)


class _controls_proxy:
    _stack = []

    def __init__(self, ctrl, context, keys=None):
        self.ctrl = ctrl
        if keys is None:
            self.keys = []
        else:
            self.keys = keys
        if context:
            self.__enter__()

    def __enter__(self):
        self._stack.append(self)

    def __exit__(self, exc_type, exc_value, traceback):
        self._stack.remove(self)


def gogogo_controls(
    kwargs,
    controls,
    display_controls,
    slider_formats,
    play_buttons,
    extra_controls=None,
    allow_dupes=False,
):
    # check if we're in a controls context manager
    if len(_controls_proxy._stack) > 0:
        ctrl_context = _controls_proxy._stack[-1]
        if extra_controls is None:
            extra_controls = [ctrl_context]
        else:
            extra_controls.append(ctrl_context)

    # Squash controls + extra_controls and make sure we have a unique controller object

    # first check whether we go an argument of controls = controls
    # or somethign like controls=controls['param1', 'param2']
    if isinstance(controls, Controls):
        ctrls = [controls]
        keys = list(controls.params.keys())
    elif isinstance(controls, _controls_proxy):
        ctrls = [controls.ctrl]
        keys = list(controls.keys)
    else:
        ctrls = []
        keys = []

    # loop over extra_controls to collect all of them
    if extra_controls is not None:
        for ec in extra_controls:
            if isinstance(ec, _controls_proxy):
                # e.g. indexed ctrls for scalar arg
                # or in context manager
                ctrls.append(ec.ctrl)
                keys.extend(ec.keys)
            elif isinstance(ec, Controls):
                # I don't think we should ever get here?
                # if we do not sure what should be happening with the keys
                ctrls.append(ec)
    ctrls = set(ctrls)
    keys = set(keys + list(kwargs.keys()))

    if len(ctrls) > 1:
        raise ValueError("Only one controls object may be used per function")

    # now we are garunteed to only have either no controls and must make one
    # or have a single controls object that we should add to.
    if None in ctrls or len(ctrls) == 0:
        controls = Controls(slider_formats=slider_formats, play_buttons=play_buttons, **kwargs)
        params = controls.params
        if display_controls:
            controls.display()
    else:
        controls = ctrls.pop()
        controls.add_kwargs(kwargs, slider_formats, play_buttons, allow_duplicates=allow_dupes)
        params = {k: controls.params[k] for k in keys}
    return controls, params


def _gen_f(key):
    def f(*args, **kwargs):
        return kwargs[key]

    return f


def _gen_param_excluder(added_kwargs):
    """
    Pass through all the original keys, but exclude any kwargs that we added
    manually through prep_scalar

    Parameters
    ----------
    added_kwargs : list of str

    Returns
    -------
    excluder : callable
    """

    def excluder(params, except_=None):
        """
        Parameters
        ----------
        params : dict
        except : str or list[str]
        """
        if isinstance(except_, str) or except_ is None:
            except_ = [except_]

        return {k: v for k, v in params.items() if k not in added_kwargs or k in except_}

    return excluder


def prep_scalars(kwarg_dict, **kwargs):
    """
    Process potentially scalar arguments. This allows for passing in
    slider shorthands for these arguments, and for passing indexed controls objects for them.

    Parameters
    ----------
    kwarg_dict : dict
        The kwargs passed to the calling functions
    kwargs :
        The arguments to process

    Returns
    -------
    funcs : dict
        The processed arguments. Re-assign the args as ``s = funcs['s']`` to
        make use of any autogenerated functions to connect callbacks to these
        scalar values.
    extra_ctrls : list of Controls
        Any extra controls objects that need to be used. Useful for `gogogo_controls`
    param_excluder : function
        A function to exclude any scalar kwargs we added from params. Use this in the
        update functions to avoid calling user provided functions with the kwargs that
        we added here.
    """
    extra_ctrls = []
    added_kwargs = []

    for name, arg in kwargs.items():
        if isinstance(arg, tuple) and name is not None:
            # name will be set by calling function if from ipyplot
            # this case is if given an abbreviation e.g.: `s = (0, 10)`

            kwargs[name] = _gen_f(name)

            # Modify the calling functions kwargs in place to add the arg
            kwarg_dict[name] = arg
            added_kwargs.append(name)
        elif isinstance(arg, _controls_proxy) and len(arg.keys) == 1:
            # indexed controls. e.g. ctrls['vmin']
            kwargs[name] = _gen_f(arg.keys[0])
            extra_ctrls.append(arg)

    if len(added_kwargs) == 0:
        # shortcircuit options
        def param_excluder(params, except_=None):
            return params

    else:
        param_excluder = _gen_param_excluder(added_kwargs)
    return kwargs, extra_ctrls, param_excluder



from matplotlib.artist import ArtistInspector
from matplotlib.collections import Collection
from matplotlib.image import AxesImage

# this is a list of options to Line2D partially taken from
# https://github.com/matplotlib/matplotlib/blob/f9d29189507cfe4121a231f6ab63539d216c37bd/lib/matplotlib/lines.py#L271
# many of these can also be made into functions
# also taken from the docstring of plt.axvline which lists all the properties of Line2D
Line2D_kwargs_list = [
    "agg_filter",
    "alpha",
    "animated",
    "antialiased",
    "aa",
    "clip_box",
    "clip_on",
    "clip_path",
    "color",
    "c",
    "contains",
    "dash_capstyle",
    "dash_joinstyle",
    "dashes",
    "data",
    "drawstyle",
    "ds",
    "figure",
    "fillstyle",
    "gid",
    "in_layout",
    "label",
    "linestyle",
    "ls",
    "linewidth",
    "lw",
    "marker",
    "markeredgecolor",
    "mec",
    "markeredgewidth",
    "mew",
    "markerfacecolor",
    "mfc",
    "markerfacecoloralt",
    "mfcalt",
    "markersize",
    "ms",
    "markevery",
    "path_effects",
    "picker",
    "pickradius",
    "rasterized",
    "sketch_params",
    "snap",
    "solid_capstyle",
    "solid_joinstyle",
    "transform",
    "url",
    "visible",
    "xdata",
    "ydata",
    "zorder",
]

imshow_kwargs_list = ArtistInspector(AxesImage).get_setters()
collection_kwargs_list = ArtistInspector(Collection).get_setters()

Text_kwargs_list = [
    "agg_filter",
    "alpha",
    "animated",
    "backgroundcolor",
    "bbox",
    "clip_box",
    "clip_on",
    "clip_path",
    "color",
    "c",
    "contains",
    "figure",
    "fontfamily",
    "family",
    "fontproperties",
    "font_properties",
    "fontsize",
    "size",
    "fontstretch",
    "stretch",
    "fontstyle",
    "style",
    "fontvariant",
    "variant",
    "fontweight" "weight",
    "gid",
    "horizontalalignment" "ha",
    "in_layout",
    "label",
    "linespacing",
    "multialignment" "ma",
    "path_effects",
    "picker",
    "position",
    "rasterized",
    "rotation",
    "rotation_mode",
    "sketch_params",
    "snap",
    "text",
    "transform",
    "url",
    "usetex",
    "verticalalignment" "va",
    "visible",
    "wrap",
    "x",
    "y",
    "zorder",
]


def kwarg_popper(kwargs, mpl_kwargs):
    """
    This will not modify kwargs for you.

    Examples
    --------
    kwargs, plot_kwargs = kwarg_popper(kwargs, plot_kwargs_list)
    """
    kwargs = dict(kwargs)
    passthrough = {}
    for k in mpl_kwargs:
        if k in kwargs:
            passthrough[k] = kwargs.pop(k)
    return kwargs, passthrough



import numpy as np

from .helpers import choose_fmt_str


def choose_datetime_nonsense(arr, timeunit="m"):
    """
    Try to do something reasonable to datetimes and timedeltas.

    Parameters
    ----------
    arr : np.array
        Array with values to be formatted.
    timeunit : str, default: m
        Truncation level for datetime and timedelta axes.

    Returns
    -------
    out : np.array
        Array modified to format decently in a slider.

    """

    if np.issubdtype(arr.dtype, "datetime64"):
        out = arr.astype(f"datetime64[{timeunit}]")
    elif np.issubdtype(arr.dtype, "timedelta64"):
        out = arr.astype(f"timedelta64[{timeunit}]").astype(int)
    else:
        out = arr
    return out


def get_hs_axes(xarr, is_color_image=False, timeunit="m"):
    """
    Read the dims and coordinates from an xarray and construct the
    axes argument for hyperslicer. Called internally by hyperslicer.

    Parameters
    ----------
    xarr : xarray.DataArray
        DataArray being viewed with hyperslicer
    is_color_image : bool, default False
        Whether the individual images of the hyperstack are color images.
    timeunit : str, default "m"
        Truncation level for datetime and timedelta axes.

    Returns
    -------
    axes : list
        axes kwarg for hyperslicer

    """
    if not is_color_image:
        dims = xarr.dims[:-2]
    else:
        dims = xarr.dims[:-3]
    coords_list = [choose_datetime_nonsense(xarr.coords[d].values, timeunit=timeunit) for d in dims]
    axes = zip(dims, coords_list)
    return list(axes)


def get_hs_extent(xarr, is_color_image=False):
    """
    Read the "YX" coordinates of an xarray.DataArray to set extent of image for
    imshow.

    Parameters
    ----------
    xarr : xarray.DataArray
        DataArray being viewed with hyperslicer
    is_color_image : bool, default False
        Whether the individual images of the hyperstack are color images.

    Returns
    -------
    extent : list
        Extent argument for imshow. [d0_min, d0_max, d1_min, d1_max]

    """

    if not is_color_image:
        dims = xarr.dims[-2:]
    else:
        dims = xarr.dims[-3:-1]

    # the reversal is because imshow transposes the array it receives
    dims = dims[::-1]
    extent = []
    for d in dims:
        vals = xarr[d].values
        extent.append(vals.min())
        extent.append(vals.max())
    return extent


def get_hs_fmts(xarr, units=None, is_color_image=False):
    """
    Get appropriate slider format strings from xarray coordinates
    based the dtype of corresponding values.

    Parameters
    ----------
    xarr : xarray.DataArray
        DataArray being viewed with hyperslicer
    units : array-like
        Units to append to end of slider value. Must have the same length
        as number of non-image dimensions in xarray.
    is_color_image : bool, default False
        Whether the individual images of the hyperstack are color images.

    Returns
    -------
    fmt_strings : dict
        Slider format strings for hyperslicer (or other mpl-interactions?)
    """
    if not is_color_image:
        dims = xarr.dims[:-2]
    else:
        dims = xarr.dims[:-3]
    fmt_strs = {}
    for i, d in enumerate(dims):
        fmt_strs[d] = choose_fmt_str(xarr[d].dtype)
        if units is not None and units[i] is not None:
            try:
                fmt_strs[d] += " {}".format(units[i])
            except KeyError:
                continue
    return fmt_strs



from collections import defaultdict
from collections.abc import Callable, Iterable
from functools import partial
from numbers import Number

import matplotlib.widgets as mwidgets
import numpy as np

try:
    import ipywidgets as widgets
    from IPython.display import display as ipy_display
except ImportError:
    pass
from matplotlib import get_backend
from matplotlib.pyplot import figure, gca, gcf, ioff
from matplotlib.pyplot import sca as mpl_sca
from numpy.distutils.misc_util import is_sequence

from .widgets import RangeSlider

__all__ = [
    "sca",
    "decompose_bbox",
    "update_datalim_from_xy",
    "update_datalim_from_bbox",
    "is_jagged",
    "broadcast_to",
    "prep_broadcast",
    "broadcast_arrays",
    "broadcast_many",
    "notebook_backend",
    "callable_else_value",
    "callable_else_value_no_cast",
    "kwarg_to_ipywidget",
    "kwarg_to_mpl_widget",
    "extract_num_options",
    "changeify",
    "create_slider_format_dict",
    "gogogo_figure",
    "gogogo_display",
    "create_mpl_controls_fig",
    "eval_xy",
    "choose_fmt_str",
]


def sca(ax):
    """
    sca that won't fail if figure not managed by pyplot
    """
    try:
        mpl_sca(ax)
    except ValueError as e:
        if "not managed by pyplot" not in str(e):
            raise e


def decompose_bbox(bbox):
    return bbox.x0, bbox.y0, bbox.x1, bbox.y1


def _update_limits(ax, x0, y0, x1, y1, x0_, y0_, x1_, y1_, stretch_x, stretch_y):
    if stretch_x:
        x0 = np.min([x0, x0_])
        x1 = np.max([x1, x1_])
    else:
        x0 = x0_
        x1 = x1_

    if stretch_y:
        y0 = np.min([y0, y0_])
        y1 = np.max([y1, y1_])
    else:
        y0 = y0_
        y1 = y1_
    # now relim and always take the maximum extent
    ax.relim()
    ax.dataLim.update_from_data_xy(np.asarray([[x0, y0], [x1, y1]]), ignore=False)


def update_datalim_from_bbox(ax, bbox, stretch_x=True, stretch_y=True):
    _update_limits(ax, *decompose_bbox(ax.dataLim), *decompose_bbox(bbox), stretch_x, stretch_y)


def update_datalim_from_xy(ax, x, y, stretch_x=True, stretch_y=True):
    """
    current : ax.dataLim
    x : array
        the new x datavalues to include
    y : array
        the new y datavalues to include
    """
    # this part bc scatter not affect by relim
    # so need this to keep stretchign working for scatter
    x0_ = np.min(x)
    x1_ = np.max(x)
    y0_ = np.min(y)
    y1_ = np.max(y)
    _update_limits(ax, *decompose_bbox(ax.dataLim), x0_, y0_, x1_, y1_, stretch_x, stretch_y)


def is_jagged(seq):
    """
    checks for jaggedness up to two dimensions
    don't need more because more doesn't make any sense for this library
    need this bc numpy is unhappy about being passed jagged arrays now :(
    """
    lens = []
    if is_sequence(seq):
        for y in seq:
            if isinstance(y, Number) or isinstance(y, Callable):
                lens.append(0)
                continue
            try:
                lens.append(len(y))
            except TypeError:
                return True
        if not all(lens[0] == l for l in lens):  # noqa: E741
            return True
    return False


def prep_broadcast(arr):
    if arr is None:
        return np.atleast_1d(None)
    if is_jagged(arr):
        arr = np.asarray(arr, dtype=np.object)
    elif isinstance(arr, Number) or isinstance(arr, Callable):
        arr = np.atleast_1d(arr)
    else:
        arr = np.atleast_1d(arr)
        if np.issubdtype(arr.dtype, np.number) and arr.ndim == 1:
            arr = arr[None, :]
    return arr


def broadcast_to(arr, to_shape, names):
    """
    happily this doesn't increase memory footprint e.g:
    import sys
    xs = np.arange(5)
    print(sys.getsizeof(xs.nbytes))
    print(sys.getsizeof(np.broadcast_to(xs, (19000, xs.shape[0]))))

    gives 28 and 112. Note 112/28 != 19000
    """
    if arr.shape[0] == to_shape[0]:
        return arr

    if arr.ndim > 1:
        if arr.shape[0] == 1:
            return np.broadcast_to(arr, (to_shape[0], *arr.shape[1:]))
        else:
            raise ValueError(f"can't broadcast {names[0]} {arr.shape} onto {names[1]} {to_shape}")
    elif arr.shape[0] == 1:
        return np.broadcast_to(arr, (to_shape[0],))
    else:
        raise ValueError(f"can't broadcast {names[0]} {arr.shape} onto {names[1]} {to_shape}")


def broadcast_arrays(*args):
    """
    This is a modified version the numpy `broadcast_arrays` function
    that uses a version of _broadcast_to that only considers the first axis
    """

    shapes = [array.shape[0] for (array, name) in args]
    idx = np.argmax(shapes)
    if all([shapes[0] == s for s in shapes]):
        # case where nothing needs to be broadcasted.
        return [array for (array, name) in args]
    return [broadcast_to(array, args[idx][0].shape, [name, args[idx][1]]) for (array, name) in args]


def broadcast_many(*args):
    """
    helper to call prep_broadcast followed by broadcast arrays
    keep as a separate function to keep the idea of broadcast_arrays the same
    """
    return broadcast_arrays(*[(prep_broadcast(arg[0]), arg[1]) for arg in args])


def notebook_backend():
    """
    returns True if the backend is ipympl or nbagg, otherwise False
    """
    backend = get_backend().lower()
    if "ipympl" in backend:
        return True
    elif backend == "nbAgg".lower():
        return True
    return False


def callable_else_value(arg, params, cache=None):
    """
    returns as a numpy array
    """
    if isinstance(arg, Callable):
        if cache:
            if arg not in cache:
                cache[arg] = np.asanyarray(arg(**params))
            return cache[arg]
        else:
            return np.asanyarray(arg(**params))
    return np.asanyarray(arg)


def callable_else_value_no_cast(arg, params, cache=None):
    """
    doesn't cast to numpy. Useful when working with parametric functions that might
    return (x, y) where it's handy to check if the return is a tuple
    """
    if isinstance(arg, Callable):
        if cache:
            if arg not in cache:
                cache[arg] = arg(**params)
            return cache[arg]
        else:
            return arg(**params)
    return arg


def callable_else_value_wrapper(arg, params, cache=None):
    def f(params):
        if isinstance(arg, Callable):
            if cache:
                if arg not in cache:
                    cache[arg] = np.asanyarray(arg(**params))
                return cache[arg]
            else:
                return np.asanyarray(arg(**params))
        return np.asanyarray(arg)

    return f


def eval_xy(x_, y_, params, cache=None):
    """
    for when y requires x as an argument and either, neither or both
    of x and y may be a function. This will automatically do the param exclusion
    for 'x' and 'y'.

    Returns
    -------
    x, y
        as numpy arrays
    """
    if "x" in params:
        # passed as a scalar with a slider
        x = params["x"]
    elif isinstance(x_, Callable):
        if cache is not None:
            if x_ in cache:
                x = cache[x_]
            else:
                x = x_(**params)
                cache[x_] = x
        else:
            x = x_(**params)
    else:
        x = x_
    if "y" in params:
        # passed a scalar with a slider
        y = params["y"]
    elif isinstance(y_, Callable):
        if cache is not None:
            if y_ in cache:
                y = cache[y_]
            else:
                y = y_(x, **params)
                cache[y_] = y
        else:
            y = y_(x, **params)
    else:
        y = y_
    return np.asanyarray(x), np.asanyarray(y)


def kwarg_to_ipywidget(key, val, update, slider_format_string, play_button=None):
    """
    Parameters
    ----------
    key : str
    val : str or number or tuple, or set or array-like
        The value to be interpreted and possibly transformed into an ipywidget
    update : callable
        The function to be called when the value of the generated widget changes.
        Must accept a dictionary *change* and an array-like *values*
    slider_format_string : str
        The format string to use for slider labels
    play_button : bool or None or str, default: None
        If true and the output widget is a slider then added a play button widget
        on the left. Also accepts 'left' or 'right' to specify the play button position.

    Returns
    -------
    init_val
        The initial value of the widget.
    control
        The generated widget. This may be the raw widget or a higher level container
        widget (e.g. HBox) depending on what widget was generated. If a fixed value is
        returned then control will be *None*
    """

    control = None
    if isinstance(val, set):
        if len(val) == 1:
            val = val.pop()
            if isinstance(val, tuple):
                # want the categories to be ordered
                pass
            else:
                # fixed parameter
                return val, None
        else:
            val = list(val)

        # categorical
        if len(val) <= 3:
            selector = widgets.RadioButtons(options=val)
        else:
            selector = widgets.Select(options=val)
        selector.observe(partial(update, values=val), names="index")
        return val[0], selector
    elif isinstance(val, widgets.Widget) or isinstance(val, widgets.fixed):
        if not hasattr(val, "value"):
            raise TypeError(
                "widgets passed as parameters must have the `value` trait."
                "But the widget passed for {key} does not have a `.value` attribute"
            )
        if isinstance(val, widgets.fixed):
            return val, None
        elif (
            isinstance(val, widgets.Select)
            or isinstance(val, widgets.SelectionSlider)
            or isinstance(val, widgets.RadioButtons)
        ):
            # all the selection widget inherit a private _Selection :(
            # it looks unlikely to change but still would be nice to just check
            # if its a subclass
            val.observe(partial(update, values=val.options), names="index")
        else:
            # set values to None and hope for the best
            val.observe(partial(update, values=None), names="value")
            return val.value, val
            # val.observe(partial(update, key=key, label=None), names=["value"])
    else:
        if isinstance(val, tuple) and val[0] in ["r", "range", "rang", "rage"]:
            # also check for some reasonably easy mispellings
            if isinstance(val[1], (np.ndarray, list)):
                vals = val[1]
            else:
                vals = np.linspace(*val[1:])
            label = widgets.Label(value=str(vals[0]))
            slider = widgets.IntRangeSlider(
                value=(0, vals.size - 1), min=0, max=vals.size - 1, readout=False, description=key
            )
            widgets.dlink(
                (slider, "value"),
                (label, "value"),
                transform=lambda x: slider_format_string.format(vals[x[0]])
                + " - "
                + slider_format_string.format(vals[x[1]]),
            )
            slider.observe(partial(update, values=vals), names="value")
            controls = widgets.HBox([slider, label])
            return vals[[0, -1]], controls

        if isinstance(val, tuple) and len(val) in [2, 3]:
            # treat as an argument to linspace
            # idk if it's acceptable to overwrite kwargs like this
            # but I think at this point kwargs is just a dict like any other
            val = np.linspace(*val)
        val = np.atleast_1d(val)
        if val.ndim > 1:
            raise ValueError(f"{key} is {val.ndim}D but can only be 1D or a scalar")
        if len(val) == 1:
            # don't need to create a slider
            return val[0], None
        else:
            # params[key] = val[0]
            label = widgets.Label(value=slider_format_string.format(val[0]))
            slider = widgets.IntSlider(min=0, max=val.size - 1, readout=False, description=key)
            widgets.dlink(
                (slider, "value"),
                (label, "value"),
                transform=lambda x: slider_format_string.format(val[x]),
            )
            slider.observe(partial(update, values=val), names="value")
            if play_button is not None and play_button is not False:
                play = widgets.Play(min=0, max=val.size - 1, step=1)
                widgets.jslink((play, "value"), (slider, "value"))
                if isinstance(play_button, str) and play_button.lower() == "right":
                    control = widgets.HBox([slider, label, play])
                else:
                    control = widgets.HBox([play, slider, label])
            else:
                control = widgets.HBox([slider, label])
            return val[0], control


def extract_num_options(val):
    """
    convert a categorical to a number of options
    """
    if len(val) == 1:
        for v in val:
            if isinstance(v, tuple):
                # this looks nightmarish...
                # but i think it should always work
                # should also check if the tuple has length one here.
                # that will only be an issue if a trailing comma was used to make the tuple ('beep',)
                # but not ('beep') - the latter is not actually a tuple
                return len(v)
            else:
                return 0
    else:
        return len(val)


def changeify(val, update):
    """
    make matplotlib update functions return a dict with key 'new'.
    Do this for compatibility with ipywidgets
    """
    update({"new": val})


def changeify_radio(val, labels, update):
    r"""
    matplolib radio buttons don't keep track what index is selected. So this
    figures out what the index is
    made a whole function bc its easier to use with partial then

    There doesn't seem to be a good way to determine which one was clicked if the
    radio button has multiple identical values but that's wildly niche
    and also probably means they're doing something they shouldn't. So: ¯\_(ツ)_/¯
    """
    update({"new": labels.index(val)})


def create_mpl_controls_fig(kwargs):
    """
    Returns
    -------
    fig : matplotlib figure
    slider_height : float
        Height of sliders in figure coordinates
    radio_height : float
        Height of radio buttons in figure coordinates

    Notes
    -----
    figure out how many inches we should devote to figure of the controls
    this is a bunch of hacky nonsense
    making it involved me holding a ruler up to my monitor
    if you have a better solution I would love to hear about it :)
    - Ian 2020-08-22

    I think maybe the correct approach is to use transforms and actually specify things in inches
    - Ian 2020-09-27
    """
    init_fig = gcf()
    n_opts = 0
    n_radio = 0
    n_sliders = 0
    for key, val in kwargs.items():
        if isinstance(val, set):
            new_opts = extract_num_options(val)
            if new_opts > 0:
                n_radio += 1
                n_opts += new_opts
        elif (
            not isinstance(val, mwidgets.AxesWidget)
            and "ipywidgets" not in str(val.__class__)  # do this to avoid depending on ipywidgets
            and isinstance(val, Iterable)
            and len(val) > 1
        ):
            n_sliders += 1

    # These are roughly the sizes used in the matplotlib widget tutorial
    # https://matplotlib.org/3.2.2/gallery/widgets/slider_demo.html#sphx-glr-gallery-widgets-slider-demo-py
    slider_in = 0.15
    radio_in = 0.6 / 3
    widget_gap_in = 0.1

    widget_inches = (
        n_sliders * slider_in + n_opts * radio_in + widget_gap_in * (n_sliders + n_radio + 1) + 0.5
    )  # half an inch for margin
    fig = None
    slider_height = 0
    radio_height = 0
    gap_height = 0
    if not all(map(lambda x: isinstance(x, mwidgets.AxesWidget), kwargs.values())):
        # if the only kwargs are existing matplotlib widgets don't make a new figure
        with ioff():
            fig = figure()
        size = fig.get_size_inches()
        fig_h = widget_inches
        fig.set_size_inches(size[0], widget_inches)
        slider_height = slider_in / fig_h
        radio_height = radio_in / fig_h
        # radio
        gap_height = widget_gap_in / fig_h
    # reset the active figure - necessary to make legends behave as expected
    # maybe this should really be handled via axes? idk
    figure(init_fig.number)
    return fig, slider_height, radio_height, gap_height


def create_mpl_selection_slider(ax, label, values, slider_format_string):
    """
    creates a slider that behaves similarly to the ipywidgets selection slider
    """
    slider = mwidgets.Slider(ax, label, 0, len(values) - 1, valinit=0, valstep=1)

    def update_text(val):
        slider.valtext.set_text(slider_format_string.format(values[int(val)]))

    # make sure the initial value also gets formatted
    update_text(0)
    slider.on_changed(update_text)
    return slider


def create_mpl_range_selection_slider(ax, label, values, slider_format_string):
    """
    creates a slider that behaves similarly to the ipywidgets selection slider
    """
    slider = RangeSlider(ax, label, 0, len(values) - 1, valinit=(0, len(values) - 1), valstep=1)

    def update_text(val):
        slider.valtext.set_text(
            slider_format_string.format(values[val[0]])
            + " - "
            + slider_format_string.format(values[val[-1]])
        )

    # make sure the initial value also gets formatted
    update_text((0, len(values) - 1))
    slider.on_changed(update_text)
    return slider


def process_mpl_widget(val, update):
    """
    handle the case of a kwarg being an existing matplotlib widget.
    This needs to be separate so that the controller can call it when mixing ipywidets and
    a widget like scatter_selector without having to create a control figure
    """
    if isinstance(val, mwidgets.RadioButtons):
        # gotta set it to the zeroth index bc there's no reasonable way to determine the current value
        # the only way the current value is stored is through the color of the circles.
        # so could query that an extract but oh boy do I ever not want to
        val.set_active(0)
        cb = val.on_clicked(partial(changeify_radio, labels=val.labels, update=update))
        return val.labels[0], val, cb
    elif isinstance(val, (mwidgets.Slider, RangeSlider)):
        # potential future improvement:
        # check if valstep has been set and then try to infer the values
        # but not now, I'm trying to avoid premature optimization lest this
        # drag on forever
        cb = val.on_changed(partial(changeify, update=partial(update, values=None)))
        return val.val, val, cb
    else:
        cb = val.on_changed(partial(changeify, update=partial(update, values=None)))
        return val.val, val, cb


def kwarg_to_mpl_widget(
    fig,
    heights,
    widget_y,
    key,
    val,
    update,
    slider_format_string,
    play_button=False,
    play_button_pos="right",
):
    """
    heights : tuple
        with slider_height, radio_height, gap_height
    returns
    -------
    init_val
    widget
    cb
        the callback id
    new_y
        The widget_y to use for the next pass
    """
    slider_height, radio_height, gap_height = heights

    # widget_y = 0.05
    slider_ax = []
    radio_ax = []
    radio_buttons = []
    if isinstance(val, set):
        if len(val) == 1:
            val = val.pop()
            if isinstance(val, tuple):
                pass
            else:
                return val, None, None, widget_y
        else:
            val = list(val)

        n = len(val)
        longest_len = max(list(map(lambda x: len(list(x)), map(str, val))))
        # should probably use something based on fontsize rather that .015
        width = max(0.15, 0.015 * longest_len)
        radio_ax = fig.add_axes([0.2, 0.9 - widget_y - radio_height * n, width, radio_height * n])
        widget_y += radio_height * n + gap_height
        radio_buttons = mwidgets.RadioButtons(radio_ax, val, active=0)
        cb = radio_buttons.on_clicked(partial(changeify_radio, labels=val, update=update))
        return val[0], radio_buttons, cb, widget_y
    elif isinstance(val, mwidgets.AxesWidget):
        val, widget, cb = process_mpl_widget(val, update)
        return val, widget, cb, widget_y
    else:
        slider = None
        if isinstance(val, tuple) and val[0] in ["r", "range", "rang", "rage"]:
            if isinstance(val[1], (np.ndarray, list)):
                vals = val[1]
            else:
                vals = np.linspace(*val[1:])
            slider_ax = fig.add_axes([0.2, 0.9 - widget_y - gap_height, 0.65, slider_height])
            slider = create_mpl_range_selection_slider(slider_ax, key, vals, slider_format_string)
            cb = slider.on_changed(partial(changeify, update=partial(update, values=vals)))
            widget_y += slider_height + gap_height
            return vals[[0, -1]], slider, cb, widget_y

        if isinstance(val, tuple):
            if len(val) == 2:
                min_ = float(val[0])
                max_ = float(val[1])
                slider_ax = fig.add_axes([0.2, 0.9 - widget_y - gap_height, 0.65, slider_height])
                slider = mwidgets.Slider(slider_ax, key, min_, max_)

                def update_text(val):
                    slider.valtext.set_text(slider_format_string.format(val))

                # make sure the initial value also gets formatted
                update_text(slider.valinit)
                slider.on_changed(update_text)
                cb = slider.on_changed(partial(changeify, update=partial(update, values=None)))
                widget_y += slider_height + gap_height
                return min_, slider, cb, widget_y
            elif len(val) == 3:
                # should warn that that doesn't make sense with matplotlib sliders
                min_ = val[0]
                max_ = val[1]
                val = np.linspace(*val)
        val = np.atleast_1d(val)
        if val.ndim > 1:
            raise ValueError(f"{key} is {val.ndim}D but can only be 1D or a scalar")
        if len(val) == 1:
            # don't need to create a slider
            return val[0], None, None, widget_y
        else:
            slider_ax = fig.add_axes([0.2, 0.9 - widget_y - gap_height, 0.65, slider_height])
            slider = create_mpl_selection_slider(slider_ax, key, val, slider_format_string)
            slider.on_changed(partial(changeify, update=partial(update, values=val)))
            widget_y += slider_height + gap_height
            return val[0], slider, None, widget_y


def create_slider_format_dict(slider_format_string):
    if isinstance(slider_format_string, defaultdict):
        return slider_format_string
    elif isinstance(slider_format_string, dict) or slider_format_string is None:
        slider_format_strings = defaultdict(lambda: "{:.2f}")
        if slider_format_string is not None:
            for key, val in slider_format_string.items():
                slider_format_strings[key] = val
    elif isinstance(slider_format_string, str):

        def f():
            return slider_format_string

        slider_format_strings = defaultdict(f)
    else:
        raise ValueError(
            f"slider_format_string must be a dict or a string but it is a {type(slider_format_string)}"
        )
    return slider_format_strings


def gogogo_figure(ipympl, ax=None):
    """
    gogogo the greatest function name of all
    """
    if ax is None:
        if ipympl:
            with ioff():
                ax = gca()
                fig = ax.get_figure()
        else:
            ax = gca()
            fig = ax.get_figure()
        return fig, ax
    else:
        return ax.get_figure(), ax


def gogogo_display(ipympl, use_ipywidgets, display, controls, fig):
    if use_ipywidgets:
        controls = widgets.VBox(controls)
        if display:
            if ipympl:
                ipy_display(widgets.VBox([controls, fig.canvas]))
            else:
                # for the case of using %matplotlib qt
                # but also want ipywidgets sliders
                # ie with force_ipywidgets = True
                ipy_display(controls)
                fig.show()
    else:
        if display:
            fig.show()
            controls[0].show()
    return controls


def choose_fmt_str(dtype=None):
    """
    Choose the appropriate string formatting for different dtypes.

    Parameters
    ----------
    dtype : np.dtype
        dtype of array containing values to be formatted.

    Returns
    -------
    fmt : str
        Bracket style format string.
    """
    if np.issubdtype(dtype, "float"):
        fmt = r"{:0.2f}"

    elif np.issubdtype(dtype, "int"):
        fmt = r"{:d}"

    else:
        fmt = r"{:}"

    return fmt



"""Custom matplotlib widgets made for use in this library."""

import numpy as np
from matplotlib import cbook, ticker
from matplotlib.cbook import CallbackRegistry
from matplotlib.widgets import AxesWidget


class scatter_selector(AxesWidget):
    """
    A widget for selecting a point in a scatter plot. callback will receive (index, (x, y))
    """

    def __init__(self, ax, x, y, pickradius=5, which_button=1, **kwargs):
        """
        Create the scatter plot and selection machinery.

        Parameters
        ----------
        ax : Axes
            The Axes on which to make the scatter plot
        x, y : float or array-like, shape (n, )
            The data positions.
        pickradius : float
            Pick radius, in points.
        which_button : int, default: 1
            Where 1=left, 2=middle, 3=right

        Other Parameters
        ----------------
        **kwargs : arguments to scatter
            Other keyword arguments are passed directly to the ``ax.scatter`` command

        """
        super().__init__(ax)
        self.scatter = ax.scatter(x, y, **kwargs, picker=True)
        self.scatter.set_pickradius(pickradius)
        self._observers = CallbackRegistry()
        self._x = x
        self._y = y
        self._button = which_button
        self.connect_event("pick_event", self._on_pick)
        self._init_val()

    def _init_val(self):
        self.val = (0, (self._x[0], self._y[0]))

    def _on_pick(self, event):
        if event.mouseevent.button == self._button:
            idx = event.ind[0]
            x = self._x[idx]
            y = self._y[idx]
            self._process(idx, (x, y))

    def _process(self, idx, val):
        self._observers.process("picked", idx, val)

    def on_changed(self, func):
        """
        When a point is clicked calll *func* with the newly selected point

        Parameters
        ----------
        func : callable
            Function to call when slider is changed.
            The function must accept a (int, tuple(float, float)) as its arguments.

        Returns
        -------
        int
            Connection id (which can be used to disconnect *func*)
        """
        return self._observers.connect("picked", lambda idx, val: func(idx, val))


class scatter_selector_index(scatter_selector):
    """
    A widget for selecting a point in a scatter plot. callback will receive the index of
    the selected point as an argument.
    """

    def _init_val(self):
        self.val = 0

    def _process(self, idx, val):
        self._observers.process("picked", idx)

    def on_changed(self, func):
        """
        When a point is clicked calll *func* with the newly selected point's index
        and position as arguments.

        Parameters
        ----------
        func : callable
            Function to call when slider is changed.
            The function must accept a single int as its arguments.

        Returns
        -------
        int
            Connection id (which can be used to disconnect *func*)
        """
        return self._observers.connect("picked", lambda idx: func(idx))


class scatter_selector_value(scatter_selector):
    """
    A widget for selecting a point in a scatter plot. callbacks will receive the x,y position of
    the selected point as arguments.
    """

    def _init_val(self):
        self.val = (self._x[0], self._y[0])

    def _process(self, idx, val):
        self._observers.process("picked", val)

    def on_changed(self, func):
        """
        When a point is clicked calll *func* with the newly selected point's index
        as arguments.

        Parameters
        ----------
        func : callable
            Function to call when slider is changed.
            The function must accept a single int as its arguments.

        Returns
        -------
        int
            Connection id (which can be used to disconnect *func*)
        """
        return self._observers.connect("picked", lambda val: func(val))


# slider widgets are taken almost verbatim from https://github.com/matplotlib/matplotlib/pull/18829/files
# which was written by me - but incorporates much of the existing matplotlib slider infrastructure
class SliderBase(AxesWidget):
    def __init__(
        self, ax, orientation, closedmin, closedmax, valmin, valmax, valfmt, dragging, valstep
    ):
        if ax.name == "3d":
            raise ValueError("Sliders cannot be added to 3D Axes")

        super().__init__(ax)

        self.orientation = orientation
        self.closedmin = closedmin
        self.closedmax = closedmax
        self.valmin = valmin
        self.valmax = valmax
        self.valstep = valstep
        self.drag_active = False
        self.valfmt = valfmt

        if orientation == "vertical":
            ax.set_ylim((valmin, valmax))
            axis = ax.yaxis
        else:
            ax.set_xlim((valmin, valmax))
            axis = ax.xaxis

        self._fmt = axis.get_major_formatter()
        if not isinstance(self._fmt, ticker.ScalarFormatter):
            self._fmt = ticker.ScalarFormatter()
            self._fmt.set_axis(axis)
        self._fmt.set_useOffset(False)  # No additive offset.
        self._fmt.set_useMathText(True)  # x sign before multiplicative offset.

        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_navigate(False)
        self.connect_event("button_press_event", self._update)
        self.connect_event("button_release_event", self._update)
        if dragging:
            self.connect_event("motion_notify_event", self._update)
        self._observers = cbook.CallbackRegistry()

    def _stepped_value(self, val):
        if self.valstep:
            return self.valmin + round((val - self.valmin) / self.valstep) * self.valstep
        return val

    def disconnect(self, cid):
        """
        Remove the observer with connection id *cid*

        Parameters
        ----------
        cid : int
            Connection id of the observer to be removed
        """
        self._observers.disconnect(cid)

    def reset(self):
        """Reset the slider to the initial value"""
        if self.val != self.valinit:
            self.set_val(self.valinit)


class RangeSlider(SliderBase):
    """
    A slider representing a floating point range.

    Create a slider from *valmin* to *valmax* in axes *ax*. For the slider to
    remain responsive you must maintain a reference to it. Call
    :meth:`on_changed` to connect to the slider event.

    Attributes
    ----------
    val : tuple[float]
        Slider value.
    """

    def __init__(
        self,
        ax,
        label,
        valmin,
        valmax,
        valinit=None,
        valfmt=None,
        closedmin=True,
        closedmax=True,
        dragging=True,
        valstep=None,
        orientation="horizontal",
        **kwargs,
    ):
        """
        Parameters
        ----------
        ax : Axes
            The Axes to put the slider in.
        label : str
            Slider label.
        valmin : float
            The minimum value of the slider.
        valmax : float
            The maximum value of the slider.
        valinit : tuple[float] or None, default: None
            The initial positions of the slider. If None the initial positions
            will be at the 25th and 75th percentiles of the range.
        valfmt : str, default: None
            %-format string used to format the slider values.  If None, a
            `.ScalarFormatter` is used instead.
        closedmin : bool, default: True
            Whether the slider interval is closed on the bottom.
        closedmax : bool, default: True
            Whether the slider interval is closed on the top.
        dragging : bool, default: True
            If True the slider can be dragged by the mouse.
        valstep : float, default: None
            If given, the slider will snap to multiples of *valstep*.
        orientation : {'horizontal', 'vertical'}, default: 'horizontal'
            The orientation of the slider.

        Notes
        -----
        Additional kwargs are passed on to ``self.poly`` which is the
        `~matplotlib.patches.Rectangle` that draws the slider knob.  See the
        `.Rectangle` documentation for valid property names (``facecolor``,
        ``edgecolor``, ``alpha``, etc.).
        """
        super().__init__(
            ax, orientation, closedmin, closedmax, valmin, valmax, valfmt, dragging, valstep
        )

        self.val = valinit
        if valinit is None:
            valinit = np.array([valmin + 0.25 * valmax, valmin + 0.75 * valmax])
        else:
            valinit = self._value_in_bounds(valinit)
        self.val = valinit
        self.valinit = valinit
        if orientation == "vertical":
            self.poly = ax.axhspan(valinit[0], valinit[1], 0, 1, **kwargs)
        else:
            self.poly = ax.axvspan(valinit[0], valinit[1], 0, 1, **kwargs)

        if orientation == "vertical":
            self.label = ax.text(
                0.5,
                1.02,
                label,
                transform=ax.transAxes,
                verticalalignment="bottom",
                horizontalalignment="center",
            )

            self.valtext = ax.text(
                0.5,
                -0.02,
                self._format(valinit),
                transform=ax.transAxes,
                verticalalignment="top",
                horizontalalignment="center",
            )
        else:
            self.label = ax.text(
                -0.02,
                0.5,
                label,
                transform=ax.transAxes,
                verticalalignment="center",
                horizontalalignment="right",
            )

            self.valtext = ax.text(
                1.02,
                0.5,
                self._format(valinit),
                transform=ax.transAxes,
                verticalalignment="center",
                horizontalalignment="left",
            )

        self.set_val(valinit)

    def _min_in_bounds(self, min):
        """
        Ensure the new min value is between valmin and self.val[1]
        """
        if min <= self.valmin:
            if not self.closedmin:
                return self.val[0]
            min = self.valmin

        if min > self.val[1]:
            min = self.val[1]
        return self._stepped_value(min)

    def _max_in_bounds(self, max):
        """
        Ensure the new max value is between valmax and self.val[0]
        """
        if max >= self.valmax:
            if not self.closedmax:
                return self.val[1]
            max = self.valmax

        if max <= self.val[0]:
            max = self.val[0]
        return self._stepped_value(max)

    def _value_in_bounds(self, val):
        return (self._min_in_bounds(val[0]), self._max_in_bounds(val[1]))

    def _update_val_from_pos(self, pos):
        """
        Given a position update the *val*
        """
        idx = np.argmin(np.abs(self.val - pos))
        if idx == 0:
            val = self._min_in_bounds(pos)
            self.set_min(val)
        else:
            val = self._max_in_bounds(pos)
            self.set_max(val)

    def _update(self, event):
        """Update the slider position."""
        if self.ignore(event) or event.button != 1:
            return

        if event.name == "button_press_event" and event.inaxes == self.ax:
            self.drag_active = True
            event.canvas.grab_mouse(self.ax)

        if not self.drag_active:
            return

        elif (event.name == "button_release_event") or (
            event.name == "button_press_event" and event.inaxes != self.ax
        ):
            self.drag_active = False
            event.canvas.release_mouse(self.ax)
            return
        if self.orientation == "vertical":
            self._update_val_from_pos(event.ydata)
        else:
            self._update_val_from_pos(event.xdata)

    def _format(self, val):
        """Pretty-print *val*."""
        if self.valfmt is not None:
            return (self.valfmt % val[0], self.valfmt % val[1])
        else:
            # fmt.get_offset is actually the multiplicative factor, if any.
            _, s1, s2, _ = self._fmt.format_ticks([self.valmin, *val, self.valmax])
            # fmt.get_offset is actually the multiplicative factor, if any.
            s1 += self._fmt.get_offset()
            s2 += self._fmt.get_offset()
            # use raw string to avoid issues with backslashes from
            return rf"({s1}, {s2})"

    def set_min(self, min):
        """
        Set the lower value of the slider to *min*

        Parameters
        ----------
        min : float
        """
        self.set_val((min, self.val[1]))

    def set_max(self, max):
        """
        Set the lower value of the slider to *max*

        Parameters
        ----------
        max : float
        """
        self.set_val((self.val[0], max))

    def set_val(self, val):
        """
        Set slider value to *val*

        Parameters
        ----------
        val : tuple or arraylike of float
        """
        val = np.sort(np.asanyarray(val))
        if val.shape != (2,):
            raise ValueError(f"val must have shape (2,) but has shape {val.shape}")
        val[0] = self._min_in_bounds(val[0])
        val[1] = self._max_in_bounds(val[1])
        xy = self.poly.xy
        if self.orientation == "vertical":
            xy[0] = 0, val[0]
            xy[1] = 0, val[1]
            xy[2] = 1, val[1]
            xy[3] = 1, val[0]
            xy[4] = 0, val[0]
        else:
            xy[0] = val[0], 0
            xy[1] = val[0], 1
            xy[2] = val[1], 1
            xy[3] = val[1], 0
            xy[4] = val[0], 0
        self.poly.xy = xy
        self.valtext.set_text(self._format(val))
        if self.drawon:
            self.ax.figure.canvas.draw_idle()
        self.val = val
        if self.eventson:
            self._observers.process("changed", val)

    def on_changed(self, func):
        """
        When the slider value is changed call *func* with the new
        slider value

        Parameters
        ----------
        func : callable
            Function to call when slider is changed.
            The function must accept a numpy array with shape (2,) float
            as its argument.

        Returns
        -------
        int
            Connection id (which can be used to disconnect *func*)
        """
        return self._observers.connect("changed", func)
